from pychess.Utils.Cord import Cord
from pychess.Utils.const import (
    DROP,
    NORMAL_MOVE,
    PAWN,
    SITTUYINCHESS,
    QUEEN,
    KING,
    WHITE,
    BLACK,
    W_OOO,
    W_OO,
    B_OOO,
    B_OO,
    QUEEN_CASTLE,
    FISCHERRANDOMCHESS,
    GATINGS,
    KING_CASTLE,
    CAMBODIANCHESS,
    ENPASSANT,
    PROMOTIONS,
    CASTLE_SAN,
    C1,
    G1,
    reprSign,
    LIGHTBRIGADECHESS,
    KNIGHT,
)
from pychess.Utils.lutils.lmovegen import newMove
from .lutils import lmove


class Move:
    def __init__(self, cord0, cord1=None, board=None, promotion=None):
        """Inits a new highlevel Move object.
        The object can be initialized in the follow ways:
            Move(cord0, cord1, board, [promotionPiece])
            Move(lovLevelMoveInt)"""

        if not cord1:
            self.move = cord0
            self.flag = self.move >> 12
            self.cord0 = None if self.flag == DROP else Cord(lmove.FCORD(self.move))
            self.cord1 = Cord(lmove.TCORD(self.move))
        else:
            assert (
                cord0 is not None and cord1 is not None
            ), f"cord0={cord0}, cord1={cord1}, board={board}"
            assert board[cord0] is not None, "cord0={}, cord1={}, board={}".format(
                cord0, cord1, board
            )
            self.cord0 = cord0
            self.cord1 = cord1
            if not board:
                raise ValueError(
                    "Move needs a Board object in order to investigate flags"
                )

            self.flag = NORMAL_MOVE
            if (
                board[self.cord0].piece == PAWN
                and self.cord1.cord in board.PROMOTION_ZONE[board.board.color]
                and board.variant != SITTUYINCHESS
            ):
                if promotion is None:
                    if board.variant == LIGHTBRIGADECHESS:
                        self.flag = lmove.FLAG_PIECE(
                            QUEEN if board.board.color == WHITE else KNIGHT
                        )
                    else:
                        self.flag = lmove.FLAG_PIECE(QUEEN)
                else:
                    self.flag = lmove.FLAG_PIECE(promotion)

            elif board[self.cord0].piece == PAWN and board.variant == SITTUYINCHESS:
                if cord0 == cord1:
                    # in place promotion
                    self.flag = lmove.FLAG_PIECE(QUEEN)
                elif (
                    board[self.cord1] is None
                    and (self.cord0.cord + self.cord1.cord) % 2 == 1
                    and (
                        self.cord0.cord in board.PROMOTION_ZONE[board.board.color]
                        or board.board.pieceCount[board.color][PAWN] == 1
                    )
                ):
                    # queen move promotion
                    self.flag = lmove.FLAG_PIECE(QUEEN)

            elif board[self.cord0].piece == KING:
                if self.cord0.x - self.cord1.x == 2 and board.variant not in (
                    CAMBODIANCHESS,
                    FISCHERRANDOMCHESS,
                ):
                    self.flag = QUEEN_CASTLE if self.cord0.x == 4 else KING_CASTLE
                elif self.cord0.x - self.cord1.x == -2 and board.variant not in (
                    CAMBODIANCHESS,
                    FISCHERRANDOMCHESS,
                ):
                    self.flag = KING_CASTLE if self.cord0.x == 4 else QUEEN_CASTLE
                elif board.variant != CAMBODIANCHESS:
                    if (
                        abs(self.cord0.x - self.cord1.x) > 1 and self.cord1.x == C1
                    ) or (
                        board.board.ini_rooks[board.color][0] == self.cord1.cord
                        and (
                            (
                                board.board.color == WHITE
                                and board.board.castling & W_OOO
                            )
                            or (
                                board.board.color == BLACK
                                and board.board.castling & B_OOO
                            )
                        )
                    ):
                        self.flag = QUEEN_CASTLE
                    elif (
                        abs(self.cord0.x - self.cord1.x) > 1 and self.cord1.x == G1
                    ) or (
                        board.board.ini_rooks[board.color][1] == self.cord1.cord
                        and (
                            (board.board.color == WHITE and board.board.castling & W_OO)
                            or (
                                board.board.color == BLACK
                                and board.board.castling & B_OO
                            )
                        )
                    ):
                        self.flag = KING_CASTLE
            elif (
                board[self.cord0].piece == PAWN
                and board[self.cord1] is None
                and self.cord0.x != self.cord1.x
                and self.cord0.y != self.cord1.y
            ):
                self.flag = ENPASSANT

            self.move = newMove(self.cord0.cord, self.cord1.cord, self.flag)

    def _get_cords(self):
        return (self.cord0, self.cord1)

    cords = property(_get_cords)

    def _get_promotion(self):
        if self.flag in PROMOTIONS:
            return lmove.PROMOTE_PIECE(self.flag)
        return None

    promotion = property(_get_promotion)

    def __repr__(self):
        promotion = (
            "=" + reprSign[lmove.PROMOTE_PIECE(self.flag)]
            if self.flag in PROMOTIONS
            else ""
        )

        gate_piece = (
            reprSign[lmove.GATE_PIECE(self.flag)].lower()
            if self.flag in GATINGS
            else ""
        )
        if self.flag == DROP:
            piece = reprSign[lmove.FCORD(self.move)]
            return piece + "@" + str(self.cord1) + promotion
        else:
            return str(self.cord0) + str(self.cord1) + gate_piece + promotion

    def __eq__(self, other):
        if isinstance(other, Move):
            return self.move == other.move

    def __hash__(self):
        return hash(self.cords)

    def is_capture(self, board):
        return (
            self.flag == ENPASSANT
            or board[self.cord1] is not None
            and self.flag != QUEEN_CASTLE
            and self.flag != KING_CASTLE
        )

    def as_uci(self):
        move = "{}{}{}{}".format(
            self.cord0.cx, self.cord0.cy, self.cord1.cx, self.cord1.cy
        )
        if self.flag in PROMOTIONS:
            move += reprSign[lmove.PROMOTE_PIECE(self.flag)].lower()
        return move


# Parsers


def listToMoves(board, mstrs, type=None, validate=False, ignoreErrors=False):
    return [
        Move(move)
        for move in lmove.listToMoves(board.board, mstrs, type, validate, ignoreErrors)
    ]


def parseAny(board, algnot):
    return Move(lmove.parseAny(board.board, algnot))


def parseSAN(board, san):
    """Parse a Short/Abbreviated Algebraic Notation string"""

    return Move(lmove.parseSAN(board.board, san))


def parseLAN(board, lan):
    """Parse a Long/Expanded Algebraic Notation string"""

    return Move(lmove.parseLAN(board.board, lan))


def parseFAN(board, lan):
    """Parse a Long/Expanded Algebraic Notation string"""

    return Move(lmove.parseFAN(board.board, lan))


def parseAN(board, an):
    """Parse an Algebraic Notation string"""

    return Move(lmove.parseAN(board.board, an))


# Exporters


def listToSan(board, moves):
    return lmove.listToSan(board.board, (m.move for m in moves))


def toAN(board, move, short=False, castleNotation=CASTLE_SAN):
    """Returns a Algebraic Notation string of a move
    board should be prior to the move"""

    return lmove.toAN(
        board.board, move.move, short=short, castleNotation=castleNotation
    )


def toSAN(board, move, localRepr=False):
    """Returns a Short/Abbreviated Algebraic Notation string of a move
    The board should be prior to the move, board2 past.
    If not board2, toSAN will not test mate"""

    return lmove.toSAN(board.board, move.move, localRepr)


def toLAN(board, move):
    """Returns a Long/Expanded Algebraic Notation string of a move
    board should be prior to the move"""

    return lmove.toLAN(board.board, move.move)


def toFAN(board, move):
    """Returns a Figurine Algebraic Notation string of a move"""

    return lmove.toFAN(board.board, move.move)



import asyncio
from io import StringIO

from pychess.System.Log import log
from pychess.Utils.GameModel import GameModel
from pychess.Utils.Offer import Offer
from pychess.Utils.const import (
    REMOTE,
    DRAW,
    WHITE,
    BLACK,
    RUNNING,
    WHITEWON,
    KILLED,
    TAKEBACK_OFFER,
    WAITING_TO_START,
    BLACKWON,
    PAUSE_OFFER,
    PAUSED,
    RESUME_OFFER,
    DISCONNECTED,
    CHAT_ACTION,
    RESIGNATION,
    FLAG_CALL,
    OFFERS,
    LOCAL,
    UNFINISHED_STATES,
    ABORT_OFFER,
    ACTION_ERROR_NONE_TO_ACCEPT,
)
from pychess.Players.Human import Human
from pychess.Savers import fen as fen_loader
from pychess.ic import GAME_TYPES, TYPE_TOURNAMENT_DIRECTOR


class ICGameModel(GameModel):
    def __init__(self, connection, ficsgame, timemodel):
        assert ficsgame.game_type in GAME_TYPES.values()
        GameModel.__init__(self, timemodel, ficsgame.game_type.variant)
        self.connection = connection
        self.ficsgame = ficsgame
        self.ficsplayers = (ficsgame.wplayer, ficsgame.bplayer)
        self.gmwidg_ready = asyncio.Event()
        self.kibitz_task = None
        self.disconnected = False

        connections = self.connections
        connections[connection.bm].append(
            connection.bm.connect("boardSetup", self.onBoardSetup)
        )
        connections[connection.bm].append(
            connection.bm.connect("exGameReset", self.onExGameReset)
        )
        connections[connection.bm].append(
            connection.bm.connect("gameUndoing", self.onGameUndoing)
        )
        connections[connection.bm].append(
            connection.bm.connect("timesUpdate", self.onTimesUpdate)
        )
        connections[connection.bm].append(
            connection.bm.connect("obsGameEnded", self.onGameEnded)
        )
        connections[connection.bm].append(
            connection.bm.connect("curGameEnded", self.onGameEnded)
        )
        connections[connection.bm].append(
            connection.bm.connect("gamePaused", self.onGamePaused)
        )
        connections[connection.bm].append(
            connection.bm.connect("madeExamined", self.onMadeExamined)
        )
        connections[connection.bm].append(
            connection.bm.connect("madeUnExamined", self.onMadeUnExamined)
        )
        connections[connection.om].append(
            connection.om.connect("onActionError", self.onActionError)
        )
        connections[connection.cm].append(
            connection.cm.connect("kibitzMessage", self.onKibitzMessage)
        )
        connections[connection.cm].append(
            connection.cm.connect("whisperMessage", self.onWhisperMessage)
        )
        connections[connection.cm].append(
            connection.cm.connect("observers_received", self.onObserversReceived)
        )
        connections[connection].append(
            connection.connect("disconnected", self.onDisconnected)
        )

        rated = "rated" if ficsgame.rated else "unrated"
        # This is in the format that ficsgames.org writes these PGN headers
        ics = "ICC" if self.connection.ICC else "FICS"
        self.tags["Event"] = "{} {} {} game".format(
            ics, rated, ficsgame.game_type.fics_name
        )
        self.tags["Site"] = "chessclub.com" if self.connection.ICC else "freechess.org"

    def __repr__(self):
        string = GameModel.__repr__(self)
        string = string.replace("<GameModel", "<ICGameModel")
        fics_game = repr(self.ficsgame)
        string = string.replace(", players=", ", ficsgame=%s, players=" % fics_game)
        return string

    @property
    def display_text(self):
        text = "[ "
        if self.timed:
            text += self.timemodel.display_text + " "
        text += self.ficsgame.display_rated.lower() + " "
        if self.ficsgame.game_type.display_text:
            text += self.ficsgame.game_type.display_text + " "
        return text + "]"

    def __disconnect(self):
        if self.connections is None:
            return
        for obj in self.connections:
            # Humans need to stay connected post-game so that "GUI > Actions" works
            if isinstance(obj, Human):
                continue

            for handler_id in self.connections[obj]:
                if obj.handler_is_connected(handler_id):
                    log.debug(
                        "ICGameModel.__disconnect: object=%s handler_id=%s"
                        % (repr(obj), repr(handler_id))
                    )
                    obj.disconnect(handler_id)
        self.disconnected = True

    def ficsplayer(self, player):
        if player.ichandle == self.ficsplayers[0].name:
            return self.ficsplayers[0]
        else:
            return self.ficsplayers[1]

    @property
    def remote_player(self):
        if self.players[0].__type__ == REMOTE:
            return self.players[0]
        else:
            return self.players[1]

    @property
    def remote_ficsplayer(self):
        return self.ficsplayer(self.remote_player)

    def hasGuestPlayers(self):
        for player in self.ficsplayers:
            if player.isGuest():
                return True
        return False

    @property
    def noTD(self):
        for player in self.ficsplayers:
            if TYPE_TOURNAMENT_DIRECTOR in player.titles:
                return False
        return True

    def onExGameReset(self, bm, ficsgame):
        log.debug("ICGameModel.onExGameReset %s" % self)
        if ficsgame == self.ficsgame:
            self.__disconnect()
            self.players[0].end()
            self.players[1].end()

    def onGameUndoing(self, bm, gameno, ply):
        if gameno == self.ficsgame.gameno:
            self.undoMoves(ply)

    def onBoardSetup(self, bm, gameno, fen, wname, bname):
        log.debug(
            "ICGameModel.onBoardSetup: {} {} {} {} {}".format(
                bm, gameno, fen, wname, bname
            )
        )
        if (
            gameno != self.ficsgame.gameno
            or len(self.players) != 2
            or self.disconnected
        ):
            return

        # Set up examined game black player
        if "Black" not in self.tags or bname != self.tags["Black"]:
            self.players[BLACK].name = bname
            self.emit("players_changed")

        # Set up examined game white player
        if "White" not in self.tags or wname != self.tags["White"]:
            self.players[WHITE].name = wname
            self.emit("players_changed")

        # Set up examined game position, side to move, castling rights
        if self.boards[-1].asFen() != fen:
            if self.boards[0].asFen().split()[:2] == fen.split()[:2]:
                log.debug("ICGameModel.onBoardSetup: undoing moves %s" % self.moves)
                self.undoMoves(len(self.moves))
                self.ficsgame.move_queue.put_nowait("stm")
                return

            new_position = self.boards[-1].asFen().split()[0] != fen.split()[0]

            # side to move change
            stm_change = self.boards[-1].asFen().split()[1] != fen.split()[1]
            self.status = RUNNING
            self.loadAndStart(StringIO(fen), fen_loader, 0, -1, first_time=False)

            if new_position:
                log.debug('ICGameModel.onBoardSetup: put_nowait("fen"')
                self.ficsgame.move_queue.put_nowait("fen")
                self.emit("game_started")
            elif stm_change:
                log.debug('ICGameModel.onBoardSetup: put_nowait("stm"')
                self.ficsgame.move_queue.put_nowait("stm")

    def update_board(self, gameno, ply, curcol, lastmove, fen, wname, bname, wms, bms):
        log.debug(
            (
                "ICGameModel.update_board: id=%s self.ply=%s self.players=%s gameno=%s "
                + "wname=%s bname=%s ply=%s curcol=%s lastmove=%s fen=%s wms=%s bms=%s"
            )
            % (
                str(id(self)),
                str(self.ply),
                repr(self.players),
                str(gameno),
                str(wname),
                str(bname),
                str(ply),
                str(curcol),
                str(lastmove),
                str(fen),
                str(wms),
                str(bms),
            )
        )
        if gameno != self.ficsgame.gameno or len(self.players) < 2 or self.disconnected:
            return

        if self.timed:
            log.debug(
                "ICGameModel.update_board: id=%d self.players=%s: updating timemodel"
                % (id(self), str(self.players))
            )
            # If game end coming from helper connection before last move made
            # we have to tap() ourselves
            if self.status in (DRAW, WHITEWON, BLACKWON):
                if self.timemodel.ply < ply:
                    self.timemodel.paused = False
                    self.timemodel.tap()
                    self.timemodel.paused = True

            self.timemodel.updatePlayer(WHITE, wms / 1000.0)
            self.timemodel.updatePlayer(BLACK, bms / 1000.0)

        if ply < self.ply:
            log.debug(
                "ICGameModel.update_board: id=%d self.players=%s \
                      self.ply=%d ply=%d: TAKEBACK"
                % (id(self), str(self.players), self.ply, ply)
            )
            for offer in list(self.offers.keys()):
                if offer.type == TAKEBACK_OFFER:
                    # There can only be 1 outstanding takeback offer for both players on FICS,
                    # (a counter-offer by the offeree for a takeback for a different number of
                    # moves replaces the initial offer) so we can safely remove all of them
                    del self.offers[offer]

            if len(self.moves) >= self.ply - ply:
                self.undoMoves(self.ply - ply)
            else:
                self.status = RUNNING
                self.loadAndStart(StringIO(fen), fen_loader, 0, -1, first_time=False)
                self.emit("game_started")
                curPlayer = self.players[self.curColor]
                curPlayer.resetPosition()

        elif ply > self.ply + 1:
            log.debug(
                "ICGameModel.update_board: id=%d self.players=%s \
                      self.ply=%d ply=%d: FORWARD JUMP"
                % (id(self), str(self.players), self.ply, ply)
            )
            self.status = RUNNING
            self.loadAndStart(StringIO(fen), fen_loader, 0, -1, first_time=False)
            self.emit("game_started")
            curPlayer = self.players[self.curColor]
            curPlayer.resetPosition()

    def onTimesUpdate(self, bm, gameno, wms, bms):
        if gameno != self.ficsgame.gameno:
            return
        if self.timed:
            self.timemodel.updatePlayer(WHITE, wms / 1000.0)
            self.timemodel.updatePlayer(BLACK, bms / 1000.0)

    def onMadeExamined(self, bm, gameno):
        self.examined = True

    def onMadeUnExamined(self, bm, gameno):
        self.examined = False

    def onGameEnded(self, bm, ficsgame):
        if ficsgame == self.ficsgame and len(self.players) >= 2:
            log.debug(
                "ICGameModel.onGameEnded: self.players=%s ficsgame=%s"
                % (repr(self.players), repr(ficsgame))
            )
            self.end(ficsgame.result, ficsgame.reason)

    def setPlayers(self, players):
        GameModel.setPlayers(self, players)
        if self.players[WHITE].icrating:
            self.tags["WhiteElo"] = self.players[WHITE].icrating
        if self.players[BLACK].icrating:
            self.tags["BlackElo"] = self.players[BLACK].icrating

        if (
            self.connection.username == self.ficsplayers[WHITE].name
        ) and self.ficsplayers[WHITE].isGuest():
            self.tags["White"] += " (Player)"
            self.emit("players_changed")
        if (
            self.connection.username == self.ficsplayers[BLACK].name
        ) and self.ficsplayers[BLACK].isGuest():
            self.tags["Black"] += " (Player)"
            self.emit("players_changed")

    def onGamePaused(self, bm, gameno, paused):
        if paused:
            self.pause()
        else:
            self.resume()

        # we have to do this here rather than in acceptReceived(), because
        # sometimes FICS pauses/unpauses a game clock without telling us that the
        # original offer was "accepted"/"received", such as when one player offers
        # "pause" and the other player responds not with "accept" but "pause"
        for offer in list(self.offers.keys()):
            if offer.type in (PAUSE_OFFER, RESUME_OFFER):
                del self.offers[offer]

    def onDisconnected(self, connection):
        if self.status in (WAITING_TO_START, PAUSED, RUNNING):
            self.end(KILLED, DISCONNECTED)

    ############################################################################
    # Chat management                                                          #
    ############################################################################

    def onKibitzMessage(self, cm, name, gameno, text):
        async def coro():
            if not self.gmwidg_ready.is_set():
                await self.gmwidg_ready.wait()
            if gameno != self.ficsgame.gameno:
                return
            self.emit("message_received", name, text)

        self.kibitz_task = asyncio.create_task(coro())

    def onWhisperMessage(self, cm, name, gameno, text):
        if gameno != self.ficsgame.gameno:
            return
        self.emit("message_received", name, text)

    def onObserversReceived(self, other, gameno, observers):
        if int(gameno) != self.ficsgame.gameno:
            return
        self.emit("observers_received", observers)

    ############################################################################
    # Offer management                                                         #
    ############################################################################

    def offerReceived(self, player, offer):
        log.debug(f"ICGameModel.offerReceived: offerer={repr(player)} {offer}")
        if player == self.players[WHITE]:
            opPlayer = self.players[BLACK]
        else:
            opPlayer = self.players[WHITE]

        if offer.type == CHAT_ACTION:
            opPlayer.putMessage(offer.param)

        elif offer.type in (RESIGNATION, FLAG_CALL):
            self.connection.om.offer(offer)

        elif offer.type in OFFERS:
            if offer not in self.offers:
                log.debug(
                    "ICGameModel.offerReceived: {}.offer({})".format(
                        repr(opPlayer), offer
                    )
                )
                self.offers[offer] = player
                opPlayer.offer(offer)
            # If the offer was an update to an old one, like a new takebackvalue
            # we want to remove the old one from self.offers
            for offer_ in list(self.offers.keys()):
                if offer.type == offer_.type and offer != offer_:
                    del self.offers[offer_]

    def acceptReceived(self, player, offer):
        log.debug(f"ICGameModel.acceptReceived: accepter={repr(player)} {offer}")
        if player.__type__ == LOCAL:
            if offer not in self.offers or self.offers[offer] == player:
                player.offerError(offer, ACTION_ERROR_NONE_TO_ACCEPT)
            else:
                log.debug(
                    "ICGameModel.acceptReceived: connection.om.accept(%s)" % offer
                )
                self.connection.om.accept(offer)
                del self.offers[offer]

        # We don't handle any ServerPlayer calls here, as the fics server will
        # know automatically if he/she accepts an offer, and will simply send
        # us the result.

    def checkStatus(self):
        pass

    def onActionError(self, om, offer, error):
        self.emit("action_error", offer, error)

    #
    # End
    #

    def end(self, status, reason):
        if self.examined:
            self.connection.bm.unexamine()

        if self.status in UNFINISHED_STATES:
            self.__disconnect()

            if self.isObservationGame():
                self.connection.bm.unobserve(self.ficsgame)
            else:
                self.connection.om.offer(Offer(ABORT_OFFER))
                self.connection.om.offer(Offer(RESIGNATION))

        if status == KILLED:
            GameModel.kill(self, reason)
        else:
            GameModel.end(self, status, reason)

    def terminate(self):
        for obj in self.connections:
            for handler_id in self.connections[obj]:
                if obj.handler_is_connected(handler_id):
                    obj.disconnect(handler_id)

        self.connections = None
        GameModel.terminate(self)
        if self.kibitz_task is not None:
            self.kibitz_task.cancel()

    def goFirst(self):
        self.connection.client.run_command("backward 999")

    def goPrev(self, step=1):
        self.connection.client.run_command("backward %s" % step)

    def goNext(self, step=1):
        self.connection.client.run_command("forward %s" % step)

    def goLast(self):
        self.connection.client.run_command("forward 999")

    def backToMainLine(self):
        self.connection.client.run_command("revert")



from gi.repository import Gtk

from pychess.Utils.const import EMPTY, FEN_EMPTY, FEN_START
from pychess.Utils.Board import Board
from pychess.Utils.Cord import Cord
from pychess.widgets.BoardControl import BoardControl
from pychess.Savers.ChessFile import LoadingError
from pychess.System.prefix import addDataPrefix
from pychess.widgets import mainwindow

__title__ = _("Preview")

__icon__ = addDataPrefix("glade/panel_games.svg")

__desc__ = _("Preview panel can filter game list by current game moves")


class PreviewPanel:
    def __init__(self, persp):
        self.persp = persp
        self.filtered = False

        self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        # buttons
        toolbar = Gtk.Toolbar()

        firstButton = Gtk.ToolButton(stock_id=Gtk.STOCK_MEDIA_PREVIOUS)
        toolbar.insert(firstButton, -1)

        prevButton = Gtk.ToolButton(stock_id=Gtk.STOCK_MEDIA_REWIND)
        toolbar.insert(prevButton, -1)

        nextButton = Gtk.ToolButton(stock_id=Gtk.STOCK_MEDIA_FORWARD)
        toolbar.insert(nextButton, -1)

        lastButton = Gtk.ToolButton(stock_id=Gtk.STOCK_MEDIA_NEXT)
        toolbar.insert(lastButton, -1)

        self.filterButton = Gtk.ToggleToolButton(Gtk.STOCK_FIND)
        self.filterButton.set_tooltip_text(_("Filter game list by current game moves"))
        toolbar.insert(self.filterButton, -1)

        addButton = Gtk.ToolButton(stock_id=Gtk.STOCK_ADD)
        addButton.set_tooltip_text(_("Add sub-fen filter from position/circles"))
        toolbar.insert(addButton, -1)

        firstButton.connect("clicked", self.on_first_clicked)
        prevButton.connect("clicked", self.on_prev_clicked)
        nextButton.connect("clicked", self.on_next_clicked)
        lastButton.connect("clicked", self.on_last_clicked)

        addButton.connect("clicked", self.on_add_clicked)
        self.filterButton.connect("clicked", self.on_filter_clicked)

        tool_box = Gtk.Box()
        tool_box.pack_start(toolbar, False, False, 0)

        # board
        self.gamemodel = self.persp.gamelist.gamemodel
        self.boardcontrol = BoardControl(self.gamemodel, {}, game_preview=True)
        self.boardview = self.boardcontrol.view
        self.board = self.gamemodel.boards[self.boardview.shown].board
        self.boardview.set_size_request(170, 170)

        self.boardview.got_started = True
        self.boardview.auto_update_shown = False

        self.box.pack_start(self.boardcontrol, True, True, 0)
        self.box.pack_start(tool_box, False, True, 0)
        self.box.show_all()

        selection = self.persp.gamelist.get_selection()
        self.conid = selection.connect_after("changed", self.on_selection_changed)
        self.persp.gamelist.preview_cid = self.conid

        # force first game to show
        self.persp.gamelist.set_cursor(0)

    def on_selection_changed(self, selection):
        model, iter = selection.get_selected()
        if iter is None:
            self.gamemodel.boards = [Board(FEN_EMPTY)]
            del self.gamemodel.moves[:]
            self.boardview.shown = 0
            self.boardview.redrawCanvas()
            return

        path = self.persp.gamelist.get_model().get_path(iter)

        rec, ply = self.persp.gamelist.get_record(path)
        if rec is None:
            return

        try:
            self.persp.chessfile.loadToModel(rec, -1, self.gamemodel)
        except LoadingError as err:
            dialogue = Gtk.MessageDialog(
                mainwindow(),
                type=Gtk.MessageType.WARNING,
                buttons=Gtk.ButtonsType.OK,
                message_format=err.args[0],
            )
            if len(err.args) > 1:
                dialogue.format_secondary_text(err.args[1])
            dialogue.connect("response", lambda dialogue, a: dialogue.hide())
            dialogue.show()
        self.boardview.noAnimation = True
        self.boardview.lastMove = None
        self.boardview._shown = self.gamemodel.lowply

        if ply > 0 or self.persp.gamelist.ply > 0:
            self.boardview.shown = ply if ply > 0 else self.persp.gamelist.ply
        else:
            self.boardview.shown = self.boardview.model.ply

    def on_first_clicked(self, button):
        self.boardview.showFirst()
        if self.filtered:
            self.update_gamelist()

    def on_prev_clicked(self, button):
        self.boardview.showPrev()
        if self.filtered:
            self.update_gamelist()

    def on_next_clicked(self, button):
        self.boardview.showNext()
        if self.filtered:
            self.update_gamelist()

    def on_last_clicked(self, button):
        self.boardview.showLast()
        if self.filtered:
            self.update_gamelist()

    def on_filter_clicked(self, button):
        self.filtered = button.get_active()
        if not self.filtered:
            self.persp.filter_panel.filterButton.set_sensitive(True)
            self.boardview.showFirst()
            self.filtered = True
            self.update_gamelist()
            self.filtered = False
        else:
            self.persp.filter_panel.filterButton.set_sensitive(False)
            self.update_gamelist()

    def on_add_clicked(self, button):
        """Create sub-fen from current FEN removing pieces not marked with circles"""

        self.board = self.gamemodel.boards[self.boardview.shown].board
        board = self.board.clone()
        fen = board.asFen()

        for cord in range(64):
            kord = Cord(cord)
            if kord not in self.boardview.circles:
                board.arBoard[cord] = EMPTY

        sub_fen = board.asFen().split()[0]

        # If all pieces removed (no circles at all) use the original FEN
        if sub_fen == "8/8/8/8/8/8/8/8":
            if fen == FEN_START:
                return
            else:
                sub_fen = fen.split()[0]

        self.persp.filter_panel.add_sub_fen(sub_fen)

    def update_gamelist(self):
        if not self.filtered:
            return

        self.board = self.gamemodel.boards[self.boardview.shown].board

        self.persp.gamelist.ply = self.board.plyCount
        self.persp.chessfile.set_fen_filter(self.board.asFen())
        self.persp.gamelist.load_games()



from gi.repository import Gtk

from pychess.Utils.Piece import Piece
from pychess.Utils.const import (
    SUICIDECHESS,
    GIVEAWAYCHESS,
    SITTUYINCHESS,
    SCHESS,
    WHITE,
    KING,
    QUEEN,
    ROOK,
    BISHOP,
    KNIGHT,
    HAWK,
    ELEPHANT,
)

from .PieceWidget import PieceWidget


class PromotionDialog:
    """:Description: A popup dialog that allows you to select form a set of pieces the exchange
    for a pawn through the promotion rule
    """

    def __init__(self, variant):
        from .gamewidget import getWidgets

        self.widgets = getWidgets()
        self.dialog = self.widgets["promotionDialog"]

        self.color = None

        if self.widgets["queenDock"].get_child() is None:
            self.widgets["queenDock"].add(PieceWidget(Piece(WHITE, QUEEN), variant))
            self.widgets["queenDock"].get_child().show()
            self.widgets["rookDock"].add(PieceWidget(Piece(WHITE, ROOK), variant))
            self.widgets["rookDock"].get_child().show()
            self.widgets["bishopDock"].add(PieceWidget(Piece(WHITE, BISHOP), variant))
            self.widgets["bishopDock"].get_child().show()
            self.widgets["knightDock"].add(PieceWidget(Piece(WHITE, KNIGHT), variant))
            self.widgets["knightDock"].get_child().show()
            self.widgets["kingDock"].add(PieceWidget(Piece(WHITE, KING), variant))
            self.widgets["kingDock"].get_child().show()
            if variant == SCHESS:
                self.widgets["hawkDock"].add(PieceWidget(Piece(WHITE, HAWK), variant))
                self.widgets["hawkDock"].get_child().show()
                self.widgets["elephantDock"].add(
                    PieceWidget(Piece(WHITE, ELEPHANT), variant)
                )
                self.widgets["elephantDock"].get_child().show()

    def setColor(self, color, variant):
        self.widgets["knightDock"].get_child().getPiece().color = color
        self.widgets["bishopDock"].get_child().getPiece().color = color
        self.widgets["rookDock"].get_child().getPiece().color = color
        self.widgets["queenDock"].get_child().getPiece().color = color
        self.widgets["kingDock"].get_child().getPiece().color = color
        if variant == SCHESS:
            self.widgets["hawkDock"].get_child().getPiece().color = color
            self.widgets["elephantDock"].get_child().getPiece().color = color

    def runAndHide(self, color, variant):
        self.setColor(color, variant)
        if variant != SUICIDECHESS and variant != GIVEAWAYCHESS:
            self.widgets["button5"].hide()

        if variant != SCHESS:
            self.widgets["button6"].hide()
            self.widgets["button7"].hide()

        if variant == SITTUYINCHESS:
            self.widgets["button4"].hide()
            self.widgets["button3"].hide()
            self.widgets["button2"].hide()

        res = self.dialog.run()
        self.dialog.hide()
        if res != Gtk.ResponseType.DELETE_EVENT:
            return [QUEEN, ROOK, BISHOP, KNIGHT, KING, HAWK, ELEPHANT][int(res)]
        return None



import asyncio
import re
import sys
import shutil

from gi.repository import GObject

from pychess.System.Log import log
from pychess.System.SubProcess import SubProcess


class Pinger(GObject.GObject):
    """The received signal contains the time it took to get response from the
    server in millisecconds. -1 means that some error occurred"""

    __gsignals__ = {
        "received": (GObject.SignalFlags.RUN_FIRST, None, (float,)),
        "error": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self, host):
        GObject.GObject.__init__(self)
        self.host = host
        self.subproc = None

        self.expression = re.compile(r"=([\d\.]+) (m?s)")

        # We need untranslated error messages in regexp search
        # below, so have to use deferred translation here
        def _(msg):
            return msg

        error = _("Destination Host Unreachable")
        self.errorExprs = (re.compile("(%s)" % error),)
        del _

        self.restartsOnDead = 3
        self.deadCount = 0

    def start(self):
        assert not self.subproc
        if sys.platform == "win32":
            args = ["-t", self.host]
        else:
            args = ["-i10", self.host]
        self.subproc = SubProcess(shutil.which("ping"), args, env={"LANG": "en"})
        asyncio.create_task(self.subproc.start())
        self.conid1 = self.subproc.connect("line", self.__handleLines)
        self.conid2 = self.subproc.connect("died", self.__handleDead)

    def __handleLines(self, subprocess, line):
        match = self.expression.search(line)
        if match:
            time, unit = match.groups()
            time = float(time)
            if unit == "s":
                time *= 1000
            self.emit("received", time)
        else:
            for expr in self.errorExprs:
                match = expr.search(line)
                if match:
                    msg = match.groups()[0]
                    self.emit("error", _(msg))

    def __handleDead(self, subprocess):
        if self.deadCount < self.restartsOnDead:
            log.warning(
                "Pinger died and restarted (%d/%d)"
                % (self.deadCount + 1, self.restartsOnDead),
                extra={"task": self.subproc.defname},
            )
            self.stop()
            self.start()
            self.deadCount += 1
        else:
            self.emit("error", _("Died"))
            self.stop()

    def stop(self):
        if not self.subproc:
            return
        # exitCode = self.subproc.gentleKill()
        self.subproc.disconnect(self.conid1)
        self.subproc.disconnect(self.conid2)
        self.subproc.terminate()
        self.subproc = None


if __name__ == "__main__":
    pinger = Pinger("google.com")

    def callback(pinger, time):
        print(time)

    pinger.connect("received", callback)
    pinger.start()
    import time

    time.sleep(5)
    pinger.stop()
    time.sleep(3)



"""The task of this module is to provide easy saving/loading of configurations
It also supports gconf like connection, so you get notices when a property
has changed."""

import sys
import os
import atexit
import builtins
import locale
from configparser import RawConfigParser

from pychess import MSYS2
from pychess.Utils.const import (
    FISCHERRANDOMCHESS,
    LOSERSCHESS,
    COUNT_OF_SOUNDS,
    SOUND_MUTE,
)
from pychess.System.Log import log
from pychess.System.prefix import addDataPrefix, addUserConfigPrefix, getDataPrefix

section = "General"
configParser = RawConfigParser(default_section=section)

for sect in ("FICS", "ICC"):
    if not configParser.has_section(sect):
        configParser.add_section(sect)

path = addUserConfigPrefix("config")
encoding = locale.getpreferredencoding()
if os.path.isfile(path):
    with open(path, encoding=encoding) as fh:
        configParser.read_file(fh)


def save_config(path=path, encoding=encoding):
    with open(path, "w", encoding=encoding) as fh:
        configParser.write(fh)


atexit.register(save_config)

if sys.platform == "win32":
    username = os.environ["USERNAME"]
else:
    from os import getuid
    from pwd import getpwuid

    userdata = getpwuid(getuid())
    realname = userdata.pw_gecos.split(",")[0]
    if realname:
        username = realname
    else:
        username = userdata.pw_name

if getattr(sys, "frozen", False) and not MSYS2:
    # pyinstaller specific!
    if hasattr(sys, "_MEIPASS"):
        base_path = sys._MEIPASS
    else:
        base_path = os.path.dirname(sys.executable)
    default_book_path = os.path.join(base_path, "pychess_book.bin")
else:
    default_book_path = os.path.join(addDataPrefix("pychess_book.bin"))

no_gettext = False


idkeyfuncs = {}
conid = 0


def notify_add(key, func, *args, section=section):
    """The signature for func must be self, client, *args, **kwargs"""
    assert isinstance(key, str)
    global conid
    idkeyfuncs[conid] = (key, func, args, section)
    conid += 1
    return conid - 1


def notify_remove(conid):
    del idkeyfuncs[conid]


if "_" not in builtins.__dir__():

    def _(text):
        return text


DEFAULTS = {
    "General": {
        "firstName": username,
        "secondName": _("Guest"),
        "showEmt": False,
        "showEval": False,
        "showBlunder": True,
        "hideTabs": False,
        "closeAll": False,
        "faceToFace": False,
        "scoreLinearScale": False,
        "showCaptured": False,
        "figuresInNotation": False,
        "moveAnimation": False,
        "noAnimation": False,
        "autoPromote": False,
        "autoRotate": False,
        "showFICSgameno": False,
        "fullAnimation": True,
        "showCords": True,
        "drawWhitePawns": True,
        "drawWhitePieces": True,
        "drawBlackPawns": True,
        "drawBlackPieces": True,
        "drawAsPawns": False,
        "drawAsWhite": False,
        "drawGrid": False,
        "activateSupportAlgorithm": False,
        "board_frame": 1,
        "board_style": 1,
        "pieceTheme": "Cburnett",
        "darkcolour": "",
        "lightcolour": "",
        "movetextFont": "FreeSerif Regular 12",
        "autoSave": True,
        "autoSavePath": os.path.expanduser("~"),
        "autoSaveFormat": "pychess",
        "saveEmt": False,
        "saveEval": False,
        "saveRatingChange": False,
        "indentPgn": False,
        "saveOwnGames": True,
        "dont_show_externals_at_startup": False,
        "max_analysis_spin": 3,
        "max_depth_spin": 20,
        "variation_threshold_spin": 50,
        "fromCurrent": True,
        "shouldWhite": True,
        "shouldBlack": True,
        "ThreatPV": False,
        "infinite_analysis": False,
        "infinite_depth": True,
        "opening_check": False,
        "opening_file_entry": default_book_path,
        "book_depth_max": 13,
        "book_exact_match": True,
        "endgame_check": False,
        "egtb_path": os.path.join(getDataPrefix()),
        "online_egtb_check": True,
        "autoCallFlag": True,
        "hint_mode": False,
        "spy_mode": False,
        "ana_combobox": 0,
        "analyzer_check": True,
        "inv_ana_combobox": 0,
        "inv_analyzer_check": False,
        "newgametasker_playercombo": 0,
        "ics_combo": 0,
        "autoLogin": False,
        "standard_toggle": True,
        "blitz_toggle": True,
        "lightning_toggle": True,
        "variant_toggle": True,
        "registered_toggle": True,
        "guest_toggle": True,
        "computer_toggle": True,
        "titled_toggle": True,
        "numberOfFingers": 0,
        "numberOfTimesLoggedInAsRegisteredUser": 0,
        "lastdifference-1": -1,
        "lastdifference-2": -1,
        "lastdifference-3": -1,
        "standard_toggle1": True,
        "blitz_toggle1": True,
        "lightning_toggle1": True,
        "variant_toggle1": True,
        "computer_toggle1": True,
        "categorycombo": 0,
        "learncombo0": 0,
        "learncombo1": 0,
        "learncombo2": 0,
        "learncombo3": 0,
        "welcome_image": addDataPrefix("glade/background.jpg"),
        "alarm_spin": 15,
        "show_tip_at_startup": True,
        "tips_seed": 0,
        "tips_index": 0,
        "dont_show_externals_at_startup": False,
        "download_timestamp": False,
        "download_chess_db": False,
        "download_scoutfish": False,
        "ngvariant1": FISCHERRANDOMCHESS,
        "ngvariant2": LOSERSCHESS,
        "useSounds": True,
        "max_log_files": 10,
        "show_sidepanels": True,
        "chat_paned_position": 100,
        "notimeRadio": 0,
        "blitzRadio": 0,
        "ngblitz min": 5,
        "ngblitz gain": 0,
        "ngblitz moves": 0,
        "rapidRadio": 0,
        "ngrapid min": 15,
        "ngrapid gain": 5,
        "ngrapid moves": 0,
        "normalRadio": 0,
        "ngnormal min": 45,
        "ngnormal gain": 15,
        "ngnormal moves": 0,
        "classicalRadio": 0,
        "playNormalRadio": 0,
        "playVariant1Radio": 0,
        "playVariant2Radio": 0,
        "ngclassical min": 3,
        "ngclassical gain": 0,
        "ngclassical moves": 40,
        "whitePlayerCombobox": 0,
        "blackPlayerCombobox": 0,
        "skillSlider1": 20,
        "skillSlider2": 20,
        "taskerSkillSlider": 20,
        "seek1Radio": 0,
        "seek2Radio": 0,
        "seek3Radio": 0,
        "challenge1Radio": 0,
        "challenge2Radio": 0,
        "challenge3Radio": 0,
    },
    "FICS": {
        "timesealCheck": True,
        "hostEntry": "freechess.org",
        "usernameEntry": "",
        "passwordEntry": "",
        "asGuestCheck": True,
    },
    "ICC": {
        "timesealCheck": True,
        "hostEntry": "chessclub.com",
        "usernameEntry": "",
        "passwordEntry": "",
        "asGuestCheck": True,
    },
}

for i in range(COUNT_OF_SOUNDS):
    DEFAULTS["General"]["soundcombo%d" % i] = SOUND_MUTE
    DEFAULTS["General"]["sounduri%d" % i] = ""


def get(key, section=section):
    try:
        default = DEFAULTS[section][key]
    except KeyError:
        # window attributes has no default values
        # print("!!! conf get() KeyError: %s %s" % (section, key))
        default = None

    try:
        value = configParser.getint(section, key, fallback=default)
        # print("... conf get %s %s: %s" % (section, key, value))
        return value
    except ValueError:
        pass

    try:
        value = configParser.getboolean(section, key, fallback=default)
        # print("... conf get %s %s: %s" % (section, key, value))
        return value
    except ValueError:
        pass

    try:
        value = configParser.getfloat(section, key, fallback=default)
        # print("... conf get %s %s: %s" % (section, key, value))
        return value
    except ValueError:
        pass

    value = configParser.get(section, key, fallback=default)
    # print("... conf get %s %s: %s" % (section, key, value))
    return value


def set(key, value, section=section):
    # print("---conf set()", section, key, value)
    try:
        configParser.set(section, key, str(value))
        with open(path, "w") as fh:
            configParser.write(fh)
    except Exception as err:
        log.error(
            "Unable to save configuration '%s'='%s' because of error: %s %s"
            % (
                repr(key),
                repr(value),
                err.__class__.__name__,
                ", ".join(str(a) for a in err.args),
            )
        )
    for key_, func, args, section_ in idkeyfuncs.values():
        if key_ == key and section_ == section:
            func(None, *args)


def hasKey(key, section=section):
    return configParser.has_option(section, key)



import collections
import inspect
import re

from pychess.System.Log import log
from pychess.ic import BLOCK_START, BLOCK_SEPARATOR, BLOCK_END, BLKCMD_PASSWORD
from pychess.ic.icc import UNIT_START, UNIT_END, DTGR_START, MY_ICC_PREFIX


class ConsoleHandler:
    def __init__(self, callback):
        self.callback = callback

    def handle(self, line):
        if line:
            self.callback(line)


class Prediction:
    def __init__(self, callback, *regexps):
        self.callback = callback
        self.name = callback.__name__
        self.regexps = []
        self.matches = ()
        self.hash = hash(callback)

        for regexp in regexps:
            self.hash ^= hash(regexp)

            if not hasattr("match", regexp):
                # FICS being fairly case insensitive, we can compile with IGNORECASE
                # to easy some expressions
                self.regexps.append(re.compile(regexp, re.IGNORECASE))

    def __hash__(self):
        return self.hash

    def __len__(self):
        return len(self.regexps)


RETURN_NO_MATCH, RETURN_MATCH, RETURN_NEED_MORE, RETURN_MATCH_END = range(4)
BL, DG, CN = range(3)


class LinePrediction(Prediction):
    def __init__(self, callback, regexp):
        Prediction.__init__(self, callback, regexp)

    async def handle(self, line):
        match = self.regexps[0].match(line)
        if match:
            self.matches = (match.string,)
            obj = self.callback(match)
            if inspect.isawaitable(obj):
                await obj
            return RETURN_MATCH
        return RETURN_NO_MATCH


class MultipleLinesPrediction(Prediction):
    def __init__(self, callback, *regexps):
        Prediction.__init__(self, callback, *regexps)
        self.matchlist = []


class NLinesPrediction(MultipleLinesPrediction):
    def __init__(self, callback, *regexps):
        MultipleLinesPrediction.__init__(self, callback, *regexps)

    async def handle(self, line):
        regexp = self.regexps[len(self.matchlist)]
        match = regexp.match(line)
        if match:
            self.matchlist.append(match)
            if len(self.matchlist) == len(self.regexps):
                self.matches = [m.string for m in self.matchlist]
                obj = self.callback(self.matchlist)
                if inspect.isawaitable(obj):
                    await obj
                del self.matchlist[:]
                return RETURN_MATCH
            return RETURN_NEED_MORE
        del self.matchlist[:]
        return RETURN_NO_MATCH


class FromPlusPrediction(MultipleLinesPrediction):
    def __init__(self, callback, regexp0, regexp1):
        MultipleLinesPrediction.__init__(self, callback, regexp0, regexp1)

    async def handle(self, line):
        if not self.matchlist:
            match = self.regexps[0].match(line)
            if match:
                self.matchlist.append(match)
                return RETURN_NEED_MORE
        else:
            match = self.regexps[1].match(line)
            if match:
                self.matchlist.append(match)
                return RETURN_NEED_MORE
            else:
                self.matches = [m.string for m in self.matchlist]
                obj = self.callback(self.matchlist)
                if inspect.isawaitable(obj):
                    await obj
                del self.matchlist[:]
                return RETURN_MATCH_END
        del self.matchlist[:]
        return RETURN_NO_MATCH


class FromABPlusPrediction(MultipleLinesPrediction):
    def __init__(self, callback, regexp0, regexp1, regexp2):
        MultipleLinesPrediction.__init__(self, callback, regexp0, regexp1, regexp2)

    async def handle(self, line):
        if not self.matchlist:
            match = self.regexps[0].match(line)
            if match:
                self.matchlist.append(match)
                return RETURN_NEED_MORE
        elif len(self.matchlist) == 1:
            match = self.regexps[1].match(line)
            if match:
                self.matchlist.append(match)
                return RETURN_NEED_MORE
        else:
            match = self.regexps[2].match(line)
            if match:
                self.matchlist.append(match)
                return RETURN_NEED_MORE
            else:
                self.matches = [m.string for m in self.matchlist]
                obj = self.callback(self.matchlist)
                if inspect.isawaitable(obj):
                    await obj
                del self.matchlist[:]
                return RETURN_MATCH_END
        del self.matchlist[:]
        return RETURN_NO_MATCH


class FromToPrediction(MultipleLinesPrediction):
    def __init__(self, callback, regexp0, regexp1):
        MultipleLinesPrediction.__init__(self, callback, regexp0, regexp1)

    async def handle(self, line):
        if not self.matchlist:
            match = self.regexps[0].match(line)
            if match:
                self.matchlist.append(match)
                return RETURN_NEED_MORE
        else:
            match = self.regexps[1].match(line)
            if match:
                self.matchlist.append(match)
                self.matches = [
                    m if isinstance(m, str) else m.string for m in self.matchlist
                ]
                obj = self.callback(self.matchlist)
                if inspect.isawaitable(obj):
                    await obj
                del self.matchlist[:]
                return RETURN_MATCH
            else:
                self.matchlist.append(line)
                return RETURN_NEED_MORE
        return RETURN_NO_MATCH


TelnetLine = collections.namedtuple("TelnetLine", ["line", "code", "code_type"])
EmptyTelnetLine = TelnetLine("", None, None)


class TelnetLines:
    def __init__(self, telnet, show_reply):
        self.telnet = telnet
        self.lines = collections.deque()
        self.block_mode = False
        self.datagram_mode = False
        self.line_prefix = None
        self.consolehandler = None
        self.show_reply = show_reply

    def appendleft(self, x):
        self.lines.appendleft(x)

    def extendleft(self, iterable):
        self.lines.extendleft(iterable)

    async def popleft(self):
        try:
            return self.lines.popleft()
        except IndexError:
            lines = await self._get_lines()
            self.lines.extend(lines)
            return self.lines.popleft() if self.lines else EmptyTelnetLine

    async def _get_lines(self):
        lines = []
        line = await self.telnet.readline()
        identifier = 0

        if line.startswith(self.line_prefix):
            line = line[len(self.line_prefix) + 1 :]

        if self.datagram_mode:
            identifier = -1
            code = 0
            unit = False
            if line.startswith(UNIT_START):
                unit = True
                unit_lines = []
                cn_code = int(line[2 : line.find(" ")])
                if MY_ICC_PREFIX in line:
                    identifier = 0
                line = await self.telnet.readline()

            if unit:
                while UNIT_END not in line:
                    if line.startswith(DTGR_START):
                        code, data = line[2:-2].split(" ", 1)
                        log.debug(
                            f"{code} {data}",
                            extra={"task": (self.telnet.name, "datagram")},
                        )
                        lines.append(TelnetLine(data, int(code), DG))
                    else:
                        if line.endswith(UNIT_END):
                            parts = line.split(UNIT_END)
                            if parts[0]:
                                unit_lines.append(parts[0])
                        else:
                            unit_lines.append(line)
                    line = await self.telnet.readline()
                if len(unit_lines) > 0:
                    text = "\n".join(unit_lines)
                    lines.append(TelnetLine(text, cn_code, CN))
                    log.debug(text, extra={"task": (self.telnet.name, "not datagram")})

        elif self.block_mode and line.startswith(BLOCK_START):
            parts = line[1:].split(BLOCK_SEPARATOR)
            if len(parts) == 3:
                identifier, code, text = parts
            elif len(parts) == 4:
                identifier, code, error_code, text = parts
            else:
                log.warning(
                    "Posing not supported yet",
                    extra={"task": (self.telnet.name, "lines")},
                )
                return lines
            code = int(code)
            identifier = int(identifier)
            if text:
                line = text
            else:
                line = await self.telnet.readline()

            while not line.endswith(BLOCK_END):
                lines.append(TelnetLine(line, code, BL))
                line = await self.telnet.readline()
            lines.append(TelnetLine(line[:-1], code, BL))

            if code != BLKCMD_PASSWORD:
                log.debug(
                    "%s %s %s"
                    % (
                        identifier,
                        code,
                        "\n".join(line.line for line in lines).strip(),
                    ),
                    extra={"task": (self.telnet.name, "command_reply")},
                )
        else:
            code = 0
            lines.append(TelnetLine(line, None, None))

        if self.consolehandler:
            if identifier == 0 or identifier in self.show_reply:
                self.consolehandler.handle(lines)
                # self.show_reply.discard(identifier)

        return lines


class PredictionsTelnet:
    def __init__(
        self, telnet, predictions, reply_cmd_dict, replay_dg_dict, replay_cn_dict
    ):
        self.telnet = telnet
        self.predictions = predictions
        self.reply_cmd_dict = reply_cmd_dict
        self.replay_dg_dict = replay_dg_dict
        self.replay_cn_dict = replay_cn_dict
        self.show_reply = set()
        self.lines = TelnetLines(telnet, self.show_reply)
        self.__command_id = 1

    async def parse(self):
        line = await self.lines.popleft()

        if not line.line:
            return  # TODO: necessary?
        # print("line.line:", line.line)
        if self.lines.datagram_mode and line.code is not None:
            if line.code_type == DG:
                callback = self.replay_dg_dict[line.code]
                callback(line.line)
                log.debug(
                    line.line, extra={"task": (self.telnet.name, callback.__name__)}
                )
                return
            elif line.code_type == CN and line.code in self.replay_cn_dict:
                callback = self.replay_cn_dict[line.code]
                callback(line.line)
                log.debug(
                    line.line, extra={"task": (self.telnet.name, callback.__name__)}
                )
                return

        predictions = (
            self.reply_cmd_dict[line.code]
            if line.code is not None and line.code in self.reply_cmd_dict
            else self.predictions
        )
        for pred in list(predictions):
            answer = await self.test_prediction(pred, line)
            # print(answer, "  parse_line: trying prediction %s for line '%s'" % (pred.name, line.line[:80]))
            if answer in (RETURN_MATCH, RETURN_MATCH_END):
                log.debug(
                    "\n".join(pred.matches),
                    extra={"task": (self.telnet.name, pred.name)},
                )
                break
        else:
            # print("  NOT MATCHED:", line.line[:80])
            if line.code != BLKCMD_PASSWORD:
                log.debug(line.line, extra={"task": (self.telnet.name, "nonmatched")})

    async def test_prediction(self, prediction, line):
        lines = []
        answer = await prediction.handle(line.line)
        while answer is RETURN_NEED_MORE:
            line = await self.lines.popleft()
            lines.append(line)
            answer = await prediction.handle(line.line)

        if lines and answer not in (RETURN_MATCH, RETURN_MATCH_END):
            self.lines.extendleft(reversed(lines))
        elif answer is RETURN_MATCH_END:
            self.lines.appendleft(line)  # re-test last line that didn't match

        return answer

    def run_command(self, text, show_reply=False):
        logtext = "*" * len(text) if self.telnet.sensitive else text
        log.debug(logtext, extra={"task": (self.telnet.name, "run_command")})
        if self.lines.block_mode:
            # TODO: reuse id after command reply handled
            self.__command_id += 1
            text = f"{self.__command_id} {text}"
            if show_reply:
                self.show_reply.add(self.__command_id)
            self.telnet.write(text)
        elif self.lines.datagram_mode:
            if show_reply:
                text = f"`{MY_ICC_PREFIX}`{text}"
            self.telnet.write("%s" % text)
        else:
            self.telnet.write("%s" % text)

    def cancel(self):
        self.run_command("quit")
        self.telnet.cancel()

    def close(self):
        # save played game (if there is any) if no moves made
        self.run_command("abort")
        self.run_command("quit")
        self.telnet.close()



from pychess.System.Log import log
from pychess.System.prefix import addDataPrefix
from pychess.Utils.const import LOCAL
from pychess.widgets.ChatView import ChatView
from pychess.ic.ICGameModel import ICGameModel
from pychess.ic.icc import DG_PLAYERS_IN_MY_GAME

__title__ = _("Chat")

__icon__ = addDataPrefix("glade/panel_chat.svg")

__desc__ = _(
    "The chat panel lets you communicate with your opponent during the game, assuming he or she is interested"
)


class Sidepanel:
    def load(self, gmwidg):
        self.gamemodel = gmwidg.gamemodel
        self.player_cid = None
        self.model_cids = [
            self.gamemodel.connect("game_started", self.onGameStarted),
            self.gamemodel.connect_after("game_terminated", self.on_game_terminated),
        ]
        self.chatView = ChatView(self.gamemodel)
        self.chatView.disable("Waiting for game to load")
        self.chatview_cid = self.chatView.connect("messageTyped", self.onMessageSent)
        if isinstance(self.gamemodel, ICGameModel):
            self.model_cids.append(
                self.gamemodel.connect(
                    "observers_received", self.chatView.update_observers
                )
            )
            self.model_cids.append(
                self.gamemodel.connect("message_received", self.onICMessageReieved)
            )
        return self.chatView

    def on_game_terminated(self, model):
        self.chatView.disconnect(self.chatview_cid)
        if (
            hasattr(self, "player")
            and hasattr(self, "player_cid")
            and not self.gamemodel.examined
        ):
            self.player.disconnect(self.player_cid)
        for cid in self.model_cids:
            self.gamemodel.disconnect(cid)

    def onGameStarted(self, gamemodel):
        if gamemodel.examined:
            if gamemodel.players[0].name == gamemodel.connection.username:
                self.player = gamemodel.players[0]
                self.opplayer = gamemodel.players[1]
            else:
                self.player = gamemodel.players[1]
                self.opplayer = gamemodel.players[0]
        elif gamemodel.isObservationGame():
            # no local player but enable chat to send/receive whisper/kibitz
            pass
        elif gamemodel.players[0].__type__ == LOCAL:
            self.player = gamemodel.players[0]
            self.opplayer = gamemodel.players[1]
            if gamemodel.players[1].__type__ == LOCAL:
                log.warning("Chatpanel loaded with two local players")
        elif gamemodel.players[1].__type__ == LOCAL:
            self.player = gamemodel.players[1]
            self.opplayer = gamemodel.players[0]
        else:
            log.info("Chatpanel loaded with no local players")
            self.chatView.hide()

        if isinstance(gamemodel, ICGameModel):
            if gamemodel.connection.ICC:
                gamemodel.connection.client.run_command(
                    "set-2 %s 1" % DG_PLAYERS_IN_MY_GAME
                )
            else:
                allob = "allob " + str(gamemodel.ficsgame.gameno)
                gamemodel.connection.client.run_command(allob)

        if (
            hasattr(self, "player")
            and not gamemodel.examined
            and self.player_cid is None
        ):
            self.player_cid = self.player.connect(
                "messageReceived", self.onMessageRecieved
            )

        self.chatView.enable()

    def onMessageRecieved(self, player, text):
        sender = (
            "pychessbot" if player.gamemodel.offline_lecture else repr(self.opplayer)
        )
        self.chatView.addMessage(sender, text)

    def onICMessageReieved(self, icgamemodel, player, text):
        self.chatView.addMessage(player, text)
        # emit an allob <gameno> to FICS
        if not icgamemodel.connection.ICC:
            allob = "allob " + str(icgamemodel.ficsgame.gameno)
            icgamemodel.connection.client.run_command(allob)

    def onMessageSent(self, chatView, text):
        if hasattr(self, "player") or self.gamemodel.examined:
            if text.startswith("# "):
                text = text[2:]
                self.gamemodel.connection.cm.whisper(text)
            elif text.startswith("whisper "):
                text = text[8:]
                self.gamemodel.connection.cm.whisper(text)
            else:
                if not hasattr(self, "player"):
                    if (
                        self.gamemodel.players[0].name
                        == self.gamemodel.connection.username
                    ):
                        self.player = self.gamemodel.players[0]
                        self.opplayer = self.gamemodel.players[1]
                    else:
                        self.player = self.gamemodel.players[1]
                        self.opplayer = self.gamemodel.players[0]

                if self.gamemodel.examined:
                    self.opplayer.putMessage(text)
                else:
                    self.player.sendMessage(text)
                self.chatView.addMessage(repr(self.player), text)
        else:
            self.gamemodel.connection.cm.whisper(text)



import asyncio

from gi.repository import GObject

from pychess.System.Log import log
from pychess.Utils.const import (
    SHUFFLECHESS,
    WILDCASTLESHUFFLECHESS,
    FISCHERRANDOMCHESS,
    RANDOMCHESS,
    ASYMMETRICRANDOMCHESS,
    FEN_START,
    WHITE,
    reprResult,
)
from pychess.ic.FICSObjects import FICSGame, FICSBoard, FICSPlayer
from pychess.ic.managers.BoardManager import BoardManager, parse_reason
from pychess.ic import (
    IC_POS_OBSERVING,
    GAME_TYPES,
    IC_STATUS_PLAYING,
    IC_POS_EXAMINATING,
)
from pychess.ic.icc import (
    DG_POSITION_BEGIN,
    DG_SEND_MOVES,
    DG_MOVE_ALGEBRAIC,
    DG_MOVE_SMITH,
    DG_MOVE_TIME,
    DG_MOVE_CLOCK,
    DG_MY_GAME_STARTED,
    DG_MY_GAME_ENDED,
    DG_STARTED_OBSERVING,
    DG_MY_GAME_RESULT,
    DG_STOP_OBSERVING,
    DG_IS_VARIATION,
    DG_ISOLATED_BOARD,
    CN_SPGN,
    DG_BACKWARD,
    DG_MOVE_LAG,
    DG_KNOWS_FISCHER_RANDOM,
    DG_SET_BOARD,
    DG_GAME_MESSAGE,
    DG_TAKEBACK,
)


class ICCBoardManager(BoardManager):
    queued_send_moves = {}

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection

        self.connection.expect_dg_line(DG_MY_GAME_STARTED, self.on_icc_my_game_started)
        self.connection.expect_dg_line(
            DG_STARTED_OBSERVING, self.on_icc_started_observing
        )
        self.connection.expect_dg_line(DG_STOP_OBSERVING, self.on_icc_stop_observing)
        self.connection.expect_dg_line(DG_MY_GAME_RESULT, self.on_icc_my_game_result)
        self.connection.expect_dg_line(DG_MY_GAME_ENDED, self.on_icc_my_game_ended)

        self.connection.expect_dg_line(DG_POSITION_BEGIN, self.on_icc_position_begin)
        self.connection.expect_dg_line(DG_SEND_MOVES, self.on_icc_send_moves)
        self.connection.expect_dg_line(DG_ISOLATED_BOARD, self.on_icc_isolated_board)
        self.connection.expect_dg_line(DG_MOVE_LAG, self.on_icc_move_lag)

        # both DG doing undo x moves in gamemodel
        self.connection.expect_dg_line(DG_BACKWARD, self.on_icc_back)
        self.connection.expect_dg_line(DG_TAKEBACK, self.on_icc_back)

        self.connection.expect_dg_line(
            DG_KNOWS_FISCHER_RANDOM, self.on_icc_knows_fischer_random
        )
        self.connection.expect_dg_line(DG_SET_BOARD, self.on_icc_set_board)
        self.connection.expect_dg_line(DG_GAME_MESSAGE, self.on_icc_game_message)

        self.connection.expect_cn_line(CN_SPGN, self.on_icc_spgn)

        self.queuedEmits = {}
        self.gamemodelStartedEvents = {}
        self.theGameImPlaying = None
        self.my_game_info = None
        self.gamesImObserving = {}

        # on observe game start, it stores number of moves we expect
        self.moves_to_go = None

        self.connection.client.run_command("set-2 %s 1" % DG_MY_GAME_STARTED)
        self.connection.client.run_command("set-2 %s 1" % DG_STARTED_OBSERVING)
        self.connection.client.run_command("set-2 %s 1" % DG_STOP_OBSERVING)
        self.connection.client.run_command("set-2 %s 1" % DG_MY_GAME_RESULT)
        self.connection.client.run_command("set-2 %s 1" % DG_MY_GAME_ENDED)

        self.connection.client.run_command("set-2 %s 1" % DG_MOVE_ALGEBRAIC)
        self.connection.client.run_command("set-2 %s 1" % DG_MOVE_SMITH)
        self.connection.client.run_command("set-2 %s 1" % DG_MOVE_TIME)
        self.connection.client.run_command("set-2 %s 1" % DG_MOVE_CLOCK)
        self.connection.client.run_command("set-2 %s 1" % DG_POSITION_BEGIN)
        self.connection.client.run_command("set-2 %s 0" % DG_IS_VARIATION)
        self.connection.client.run_command("set-2 %s 0" % DG_MOVE_LAG)

        self.connection.client.run_command("set-2 %s 1" % DG_SEND_MOVES)
        self.connection.client.run_command("set-2 %s 1" % DG_ISOLATED_BOARD)
        self.connection.client.run_command("set-2 %s 1" % DG_BACKWARD)
        self.connection.client.run_command("set-2 %s 1" % DG_TAKEBACK)

        self.connection.client.run_command("set-2 %s 1" % DG_KNOWS_FISCHER_RANDOM)
        self.connection.client.run_command("set-2 %s 1" % DG_SET_BOARD)
        self.connection.client.run_command("set-2 %s 1" % DG_GAME_MESSAGE)

        self.connection.client.run_command("set style 13")

        # don't unobserve games when we start a new game, or new observe
        self.connection.client.run_command("set unobserve 0")
        self.connection.lvm.autoFlagNotify()

    def on_icc_spgn(self, data):
        log.debug("DG_SPGN %s" % data)
        if self.connection.query_game is None:
            return

        game = self.connection.query_game
        game.board = FICSBoard(0, 0, pgn=data)
        game = self.connection.games.get(game)

        self.emit("archiveGamePreview", game)

    def on_icc_game_message(self, data):
        log.debug("DG_GAME_MESSAGE %s" % data)
        return

    def on_icc_knows_fischer_random(self, data):
        log.debug("DG_KNOWS_FISCHER_RANDOM %s" % data)

    def on_icc_set_board(self, data):
        log.debug("DG_SET_BOARD %s" % data)

    def on_icc_isolated_board(self, data):
        log.debug("DG_ISOLATED_BOARD %s" % data)

    def on_icc_move_lag(self, data):
        log.debug("DG_MOVE_LAG %s" % data)

    def on_icc_back(self, data):
        log.debug("DG_ICC_BACKWARD or DG_TAKEBACK %s" % data)
        # gamenumber backup-count
        gameno, backup_count = data.split()
        gameno = int(gameno)
        backup_count = int(backup_count)
        try:
            game = self.connection.games.get_game_by_gameno(gameno)
        except KeyError:
            return

        if game == self.theGameImPlaying or game in self.gamesImObserving:
            curcol, ply, wms, bms = self.my_game_info
            for i in range(backup_count):
                ply -= 1
                curcol = 1 - curcol
            self.my_game_info = (curcol, ply, wms, bms)

            self.emit("gameUndoing", gameno, backup_count)

    def on_icc_my_game_started(self, data):
        log.debug("DG_MY_GAME_STARTED %s" % data)
        # gamenumber whitename blackname wild-number rating-type rated
        # white-initial white-increment black-initial black-increment
        # played-game {ex-string} white-rating black-rating game-id
        # white-titles black-titles irregular-legality irregular-semantics
        # uses-plunkers fancy-timecontrol promote-to-king
        # 685 Salsicha MaxiBomb 0 Blitz 1 3 0 3 0 1 {} 2147 2197 1729752694 {} {} 0 0 0 {} 0
        # 259 Rikikilord ARMH 0 Blitz 1 2 12 2 12 0 {Ex: Rikikilord 0} 1532 1406 1729752286 {} {} 0 0 0 {} 0
        (
            gameno,
            wname,
            bname,
            wild,
            rtype,
            rated,
            wmin,
            winc,
            bmin,
            binc,
            played_game,
            rest,
        ) = data.split(" ", 11)

        parts = rest.split("}", 1)[1].split()
        wrating = int(parts[0])
        brating = int(parts[1])

        gameno = int(gameno)
        wplayer = self.connection.players.get(wname)
        bplayer = self.connection.players.get(bname)
        if int(wild) > 0:
            game_type = GAME_TYPES["w%s" % wild]
        else:
            game_type = GAME_TYPES[rtype.lower()]
        log.debug("DG_MY_GAME_STARTED game type is %s" % game_type)
        for player, rating in ((wplayer, wrating), (bplayer, brating)):
            try:
                if player.ratings[game_type.rating_type] != rating:
                    player.ratings[game_type.rating_type] = rating
                    player.emit("ratings_changed", game_type.rating_type, player)
            except IndexError:
                log.debug(
                    "!!! game_type.rating_type {} is out of range in player.ratings {}".format(
                        game_type.rating_type, player.ratings
                    )
                )

        wms = bms = int(wmin) * 60 * 1000 + int(winc) * 1000
        # TODO: maybe use DG_POSITION_BEGIN2 and DG_PAST_MOVE ?
        fen = FEN_START

        game = FICSGame(
            wplayer,
            bplayer,
            gameno=gameno,
            rated=rated == "1",
            game_type=game_type,
            minutes=int(wmin),
            inc=int(winc),
            board=FICSBoard(wms, bms, fen=fen),
        )

        if self.connection.examined_game is not None:
            pgnHead = [
                ("Event", "ICC examined game"),
                ("Site", "chessclub.com"),
                ("White", wname),
                ("Black", bname),
                ("Result", "*"),
                ("SetUp", "1"),
                ("FEN", fen),
            ]
            pgn = "\n".join(['[%s "%s"]' % line for line in pgnHead]) + "\n*\n"

            game.relation = IC_POS_EXAMINATING
            game.game_type = GAME_TYPES["examined"]
            game.board.pgn = pgn

        game = self.connection.games.get(game)

        for player in (wplayer, bplayer):
            if player.status != IC_STATUS_PLAYING:
                player.status = IC_STATUS_PLAYING
            if player.game != game:
                player.game = game

        self.theGameImPlaying = game
        self.my_game_info = (WHITE, 0, wms, bms)
        self.gamemodelStartedEvents[gameno] = asyncio.Event()
        self.connection.client.run_command("follow")

        if self.connection.examined_game is not None:
            self.emit("exGameCreated", game)
        else:
            if int(wild) in (1, 2, 3, 4, 20, 21, 22):
                # several wild variant (including loadfen/loadgame) need
                # a starting FEN coming in a DG_POSITION_BEGIN datgram
                log.debug("wild20 is waiting for a starting FEN...")
            else:
                self.emit("playGameCreated", game)

    def on_icc_started_observing(self, data):
        log.debug("DG_STARTED_OBSERVING %s" % data)
        (
            gameno,
            wname,
            bname,
            wild,
            rtype,
            rated,
            wmin,
            winc,
            bmin,
            binc,
            played_game,
            rest,
        ) = data.split(" ", 11)

        parts = rest.split("}", 1)[1].split()
        wrating = int(parts[0])
        brating = int(parts[1])

        gameno = int(gameno)
        wplayer = self.connection.players.get(wname)
        bplayer = self.connection.players.get(bname)
        game_type = GAME_TYPES[rtype.lower()]

        for player, rating in ((wplayer, wrating), (bplayer, brating)):
            if player.ratings[game_type.rating_type] != rating:
                player.ratings[game_type.rating_type] = rating
                player.emit("ratings_changed", game_type.rating_type, player)

        relation = IC_POS_OBSERVING
        wms = bms = int(wmin) * 60 * 1000 + int(winc) * 1000

        game = FICSGame(
            wplayer,
            bplayer,
            gameno=gameno,
            rated=rated == "1",
            game_type=game_type,
            minutes=int(wmin),
            inc=int(winc),
            relation=relation,
        )

        game = self.connection.games.get(game, emit=False)

        self.gamesImObserving[game] = (WHITE, 0, wms, bms)
        self.queued_send_moves[game.gameno] = []
        self.queuedEmits[game.gameno] = []
        self.gamemodelStartedEvents[game.gameno] = asyncio.Event()

    def on_icc_stop_observing(self, data):
        log.debug("DG_STOP_OBSERVING %s" % data)
        gameno = int(data.split()[0])
        try:
            del self.gamemodelStartedEvents[gameno]
            game = self.connection.games.get_game_by_gameno(gameno)
        except KeyError:
            return
        self.emit("obsGameUnobserved", game)

    def on_icc_my_game_ended(self, data):
        log.debug("DG_MY_GAME_ENDED %s" % data)

    def on_icc_my_game_result(self, data):
        log.debug("DG_MY_GAME_RESULT %s" % data)
        # gamenumber become-examined game_result_code score_string2 description-string ECO
        # 1242 1 Res 1-0 {Black resigns} {D89}
        parts = data.split(" ", 4)
        gameno, ex, result_code, result, rest = parts
        gameno = int(gameno)
        comment, rest = rest[2:].split("}", 1)

        try:
            game = self.connection.games.get_game_by_gameno(gameno)
        except KeyError:
            return
        wname = game.wplayer.name
        bname = game.bplayer.name

        result, reason = parse_reason(reprResult.index(result), comment, wname=wname)

        try:
            wplayer = self.connection.players.get(wname)
            wplayer.restore_previous_status()
            # no status update will be sent by
            # FICS if the player doesn't become available, so we restore
            # previous status first (not necessarily true, but the best guess)
        except KeyError:
            log.debug("%s not in self.connections.players - creating" % wname)
            wplayer = FICSPlayer(wname)

        try:
            bplayer = self.connection.players.get(bname)
            bplayer.restore_previous_status()
        except KeyError:
            log.debug("%s not in self.connections.players - creating" % bname)
            bplayer = FICSPlayer(bname)

        game = FICSGame(
            wplayer, bplayer, gameno=int(gameno), result=result, reason=reason
        )
        if wplayer.game is not None:
            game.rated = wplayer.game.rated
        game = self.connection.games.get(game, emit=False)
        self.connection.games.game_ended(game)
        # Do this last to give anybody connected to the game's signals a chance
        # to disconnect from them first
        wplayer.game = None
        bplayer.game = None

    def on_icc_position_begin(self, data):
        log.debug("DG_POSITION_BEGIN %s" % data)
        # gamenumber {initial-FEN} nmoves-to-follow
        gameno, right_part = data.split("{")
        gameno = int(gameno)

        try:
            game = self.connection.games.get_game_by_gameno(gameno)
        except KeyError:
            return

        fen, moves_to_go = right_part.split("}")
        self.moves_to_go = int(moves_to_go)

        if fen != FEN_START and (
            game.game_type == GAME_TYPES["w20"]
            or game.game_type == GAME_TYPES["w21"]
            or game.game_type.variant_type
            in (
                SHUFFLECHESS,
                WILDCASTLESHUFFLECHESS,
                FISCHERRANDOMCHESS,
                RANDOMCHESS,
                ASYMMETRICRANDOMCHESS,
            )
        ):
            # a starting fen coming in a DG_POSITION_BEGIN datagram
            log.debug("wild20 got a starting FEN %s" % fen)
            game.board.fen = fen
            self.emit("playGameCreated", game)

    def on_icc_send_moves(self, data):
        log.debug("DG_SEND_MOVES %s" % data)
        # gamenumber algebraic-move smith-move time clock
        send_moves = data
        gameno, san_move, alg_move, time, clock = send_moves.split()
        gameno = int(gameno)

        try:
            game = self.connection.games.get_game_by_gameno(gameno)
        except KeyError:
            log.debug("Game %s is not in self.connection.games" % gameno)
            return

        fen = ""

        if game == self.theGameImPlaying:
            curcol, ply, wms, bms = self.my_game_info
        else:
            curcol, ply, wms, bms = self.gamesImObserving[game]

        if curcol == WHITE:
            wms = int(clock) * 1000
        else:
            bms = int(clock) * 1000

        ply += 1
        curcol = 1 - curcol

        if game == self.theGameImPlaying:
            self.my_game_info = (curcol, ply, wms, bms)
        else:
            self.gamesImObserving[game] = (curcol, ply, wms, bms)

        if gameno in self.queued_send_moves:
            self.queued_send_moves[gameno].append(send_moves)
            if (
                self.moves_to_go
                and len(self.queued_send_moves[gameno]) < self.moves_to_go
            ):
                return

        if self.moves_to_go == 0 or self.moves_to_go is None:
            log.debug("put san_move to move_queue %s" % san_move)
            game.move_queue.put_nowait(
                (
                    gameno,
                    ply,
                    curcol,
                    san_move,
                    fen,
                    game.wplayer.name,
                    game.bplayer.name,
                    wms,
                    bms,
                )
            )
            self.emit("timesUpdate", gameno, wms, bms)
        else:
            if game.gameno not in self.gamemodelStartedEvents:
                return
            if game.gameno not in self.queuedEmits:
                return

            pgnHead = [
                (
                    "Event",
                    "ICC {} {} game".format(
                        game.display_rated.lower(), game.game_type.fics_name
                    ),
                ),
                ("Site", "chessclub.com"),
                ("White", game.wplayer.name),
                ("Black", game.bplayer.name),
                ("Result", "*"),
                ("TimeControl", "%d+%d" % (game.minutes * 60, game.inc)),
            ]
            wrating = game.wplayer.ratings[game.game_type.rating_type]
            brating = game.bplayer.ratings[game.game_type.rating_type]
            if wrating != 0:
                pgnHead += [("WhiteElo", wrating)]
            if brating != 0:
                pgnHead += [("BlackElo", brating)]

            pgn = "\n".join(['[%s "%s"]' % line for line in pgnHead]) + "\n"

            moves = self.queued_send_moves[gameno]
            ply = 0
            for send_moves in moves:
                gameno_, san_move, alg_move, time, clock = send_moves.split()
                if ply % 2 == 0:
                    pgn += "%d. " % (ply // 2 + 1)
                pgn += f"{san_move} {{[%emt {time}]}} "
                ply += 1
            pgn += "*\n"
            del self.queued_send_moves[gameno]
            self.moves_to_go = None

            wms = bms = 0
            game = FICSGame(
                game.wplayer,
                game.bplayer,
                game_type=game.game_type,
                result=game.result,
                rated=game.rated,
                minutes=game.minutes,
                inc=game.inc,
                board=FICSBoard(wms, bms, pgn=pgn),
            )
            in_progress = True
            if in_progress:
                game.gameno = gameno
            else:
                if gameno is not None:
                    game.gameno = gameno
                # game.reason = reason
            game = self.connection.games.get(game, emit=False)

            self.emit("obsGameCreated", game)
            try:
                self.gamemodelStartedEvents[game.gameno].wait()
            except KeyError:
                pass

            for emit in self.queuedEmits[game.gameno]:
                emit()
            del self.queuedEmits[game.gameno]

            curcol, ply, wms, bms = self.gamesImObserving[game]
            self.emit("timesUpdate", game.gameno, wms, bms)



from gi.repository import Gtk, Gdk

from pychess.Savers.database import parseDateTag
from pychess.Database.PgnImport import dedicated_tags
from pychess.Utils.const import (
    BLACK,
    WHITE,
    DRAW,
    WHITEWON,
    BLACKWON,
    RUNNING,
    reprResult,
)
from pychess.perspectives import perspective_manager
from pychess.Utils.elo import get_elo_rating_change_str

firstRun = True

tags_store = Gtk.ListStore(str, str)


def run(widgets):
    global firstRun, tags_store

    # Data from the game
    persp = perspective_manager.get_perspective("games")
    gamemodel = persp.cur_gmwidg().gamemodel

    # Initialization
    if firstRun:
        initialize(widgets)
        firstRun = False

    # Load of the tags having a dedicated field
    for tag in dedicated_tags:
        tag_value = gamemodel.tags[tag]
        if tag_value is None:
            continue
        if tag == "Date":
            tag_value = tag_value.replace(".??", "").replace("????.", "")
        elif tag_value == "?":
            tag_value = ""
        widgets["%s_entry" % tag.lower()].set_text(tag_value)
    refresh_elo_rating_change(widgets)

    combo = widgets["result_combo"]
    acive_id = reprResult[gamemodel.status]
    combo.set_active_id(acive_id)

    # Load of the tags in the editor
    tags_store.clear()
    tlist = []
    for tag in gamemodel.tags:
        if (
            tag not in dedicated_tags
            and isinstance(gamemodel.tags[tag], str)
            and gamemodel.tags[tag]
        ):
            tlist.append(tag)
    tlist.sort()
    for tag in tlist:
        tags_store.append([tag, gamemodel.tags[tag]])

    # Show the loaded dialog
    widgets["game_info"].show()


def initialize(widgets):
    def hide_window(button, *args):
        widgets["game_info"].hide()
        return True

    def on_add_tag(button, *args):
        tv_iter = tags_store.append([_("New"), ""])
        path = tags_store.get_path(tv_iter)
        widgets["tags_treeview"].set_cursor(path)

    def on_delete_tag(button, *args):
        store, tv_iter = widgets["tags_treeview"].get_selection().get_selected()
        if tv_iter:
            store.remove(tv_iter)

    def accept_new_properties(button, *args):
        persp = perspective_manager.get_perspective("games")
        gamemodel = persp.cur_gmwidg().gamemodel

        # Remove the existing tags in string format
        for tag in list(gamemodel.tags):
            if isinstance(gamemodel.tags[tag], str):
                del gamemodel.tags[tag]

        # Copy of the tags from the dedicated fields
        for tag in dedicated_tags:
            gamemodel.tags[tag] = widgets["%s_entry" % tag.lower()].get_text()

        combo = widgets["result_combo"]
        tree_iter = combo.get_active_iter()
        if tree_iter is not None:
            model = combo.get_model()
            status = model[tree_iter][0]
            if status != gamemodel.status:
                gamemodel.status = status
                gamemodel.checkStatus()

        # Copy the extra tags from the editor
        for tag in tags_store:
            if tag[0] != "" and tag not in dedicated_tags:
                gamemodel.tags[tag[0]] = tag[1]

        widgets["game_info"].hide()

        # Apply some settings to the game model
        gamemodel.players[BLACK].setName(gamemodel.tags["Black"])
        gamemodel.players[WHITE].setName(gamemodel.tags["White"])
        gamemodel.emit("players_changed")
        return True

    # Tag editor
    def tag_edited_cb(cell, path, new_text):
        global tags_store
        tags_store[path][0] = new_text

    def value_edited_cb(cell, path, new_text):
        global tags_store
        tags_store[path][1] = new_text

    global tags_store
    tv_tags = widgets["tags_treeview"]
    tv_tags.set_model(tags_store)

    tag_renderer = Gtk.CellRendererText()
    tag_renderer.set_property("editable", True)
    tag_renderer.connect("edited", tag_edited_cb)
    tv_tags.append_column(Gtk.TreeViewColumn(_("Tag"), tag_renderer, text=0))

    value_renderer = Gtk.CellRendererText()
    value_renderer.set_property("editable", True)
    value_renderer.connect("edited", value_edited_cb)
    tv_tags.append_column(Gtk.TreeViewColumn(_("Value"), value_renderer, text=1))

    result_combo = widgets["result_combo"]
    result_store = Gtk.ListStore(int, str)
    for result in (
        (WHITEWON, "1-0"),
        (BLACKWON, "0-1"),
        (DRAW, "1/2-1/2"),
        (RUNNING, "*"),
    ):
        result_store.append(result)
    result_combo.set_model(result_store)
    result_combo.set_id_column(1)
    renderer_text = Gtk.CellRendererText()
    result_combo.pack_start(renderer_text, True)
    result_combo.add_attribute(renderer_text, "text", 1)

    # Events on the UI
    widgets["whiteelo_entry"].connect(
        "changed", lambda p: refresh_elo_rating_change(widgets)
    )
    widgets["blackelo_entry"].connect(
        "changed", lambda p: refresh_elo_rating_change(widgets)
    )
    widgets["date_button"].connect("clicked", on_pick_date, widgets["date_entry"])
    widgets["tag_add_button"].connect("clicked", on_add_tag)
    widgets["tag_delete_button"].connect("clicked", on_delete_tag)
    widgets["game_info"].connect("delete-event", hide_window)
    widgets["game_info_cancel_button"].connect("clicked", hide_window)
    widgets["game_info_ok_button"].connect("clicked", accept_new_properties)


red = Gdk.RGBA(0.643, 0, 0, 1)
green = Gdk.RGBA(0.306, 0.604, 0.024, 1)
black = Gdk.RGBA(0.0, 0.0, 0.0, 1.0)


def refresh_elo_rating_change(widgets):
    persp = perspective_manager.get_perspective("games")
    gamemodel = persp.cur_gmwidg().gamemodel

    site = gamemodel.tags["Site"]
    if site is not None and (
        "lichess.org" in site or "chessclub.com" in site or "freechess.org" in site
    ):
        # TODO : lichess takes 3 parameters per player
        widgets["w_elo_change"].set_text("")
        widgets["b_elo_change"].set_text("")
        return

    welo = widgets["whiteelo_entry"].get_text()
    belo = widgets["blackelo_entry"].get_text()

    wchange = get_elo_rating_change_str(gamemodel, WHITE, welo, belo)
    widgets["w_elo_change"].set_text(wchange)
    if wchange.startswith("+") or wchange.startswith("-"):
        widgets["w_elo_change"].override_color(
            Gtk.StateFlags.NORMAL, red if wchange.startswith("-") else green
        )
    else:
        widgets["w_elo_change"].override_color(Gtk.StateFlags.NORMAL, black)

    bchange = get_elo_rating_change_str(gamemodel, BLACK, welo, belo)
    widgets["b_elo_change"].set_text(bchange)
    if bchange.startswith("+") or bchange.startswith("-"):
        widgets["b_elo_change"].override_color(
            Gtk.StateFlags.NORMAL, red if bchange.startswith("-") else green
        )
    else:
        widgets["b_elo_change"].override_color(Gtk.StateFlags.NORMAL, black)


def on_pick_date(button, date_entry):
    # Parse the existing date
    date = date_entry.get_text()
    year, month, day = parseDateTag(date)

    # Prepare the date of the picker
    calendar = Gtk.Calendar()
    curyear, curmonth, curday = calendar.get_date()
    year = curyear if year is None else year
    month = curmonth if month is None else month - 1
    day = curday if day is None else day
    calendar.select_month(month, year)
    calendar.select_day(day)

    # Show the dialog
    dialog = Gtk.Dialog(
        _("Pick a date"),
        None,
        Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
        (
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_OK,
            Gtk.ResponseType.ACCEPT,
        ),
    )

    sw = Gtk.ScrolledWindow()
    sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
    sw.add(calendar)

    dialog.get_content_area().pack_start(sw, True, True, 0)
    dialog.resize(300, 200)
    dialog.show_all()

    response = dialog.run()
    dialog.destroy()

    if response == Gtk.ResponseType.ACCEPT:
        year, month, day = calendar.get_date()
        date_entry.set_text("%04d.%02d.%02d" % (year, month + 1, day))



import math

import cairo

from pychess.System import conf
from pychess.widgets.BoardView import BoardView, matrixAround

__label__ = _("Png image")
__ending__ = "png"
__append__ = False

SQUARE = 80


def save(file, model, position=None, flip=False):
    """Export the current position into a .png file"""

    show_cords = conf.get("showCords")
    boardview = BoardView(model)
    padding = int(SQUARE / 4) if show_cords else 0

    width = SQUARE * 8 + padding * 2
    height = SQUARE * 8 + padding * 2

    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
    context = cairo.Context(surface)

    boardview.shown = position
    boardview.square = 0 + padding, 0 + padding, SQUARE * 8, SQUARE

    if flip:
        boardview._rotation = math.pi
        boardview.matrix = cairo.Matrix.init_rotate(math.pi)

    boardview.matrix, boardview.invmatrix = matrixAround(
        boardview.matrix, width / 2.0, height / 2.0
    )
    context.transform(boardview.matrix)

    boardview.drawBoard(context, None)
    boardview.drawPieces(context, None)

    if show_cords:
        boardview.drawCords(context, None)

    surface.write_to_png(file.name)



from collections import defaultdict

from gi.repository import Gtk, GObject


POSITIONS_COUNT = 5
NORTH, EAST, SOUTH, WEST, CENTER = range(POSITIONS_COUNT)
reprPos = ("NORTH", "EAST", "SOUTH", "WEST", "CENTER")


class TabReceiver(Gtk.Alignment):
    __instances = defaultdict(list)

    def __init__(self, perspective):
        GObject.GObject.__init__(self)
        self.__instances[perspective].append(self)

    def _del(self):
        try:
            index = TabReceiver.__instances[self.perspective].index(self)
        except ValueError:
            return
        del TabReceiver.__instances[self.perspective][index]

    def getInstances(self, perspective):
        return iter(self.__instances[perspective])

    def showArrows(self):
        raise NotImplementedError

    def hideArrows(self):
        raise NotImplementedError



import asyncio

from gi.repository import GObject

from pychess.Utils.const import (
    BLACKWON,
    WHITEWON,
    DRAW,
    UNDOABLE_STATES,
    CANCELLED,
    PRACTICE_GOAL_REACHED,
    BLACK,
    WHITE,
    LECTURE,
    LESSON,
    PUZZLE,
    ENDGAME,
)
from pychess.Utils.GameModel import GameModel

learn2str = {
    LECTURE: "Lecture",
    PUZZLE: "Puzzle",
    LESSON: "Lesson",
    ENDGAME: "Endgame",
}

MATE, MATE_IN, DRAW_IN, EQUAL_IN, EVAL_IN, PROMOTION = range(6)


class Goal:
    def __init__(self, termination):
        self.termination = termination
        # print(termination)

        if termination.startswith("mate in"):
            self.result = MATE_IN
            self.moves = int(termination.split()[-1])
            self.cp = None
        elif termination.startswith("draw in"):
            self.result = DRAW_IN
            self.moves = int(termination.split()[-1])
            self.cp = None
        elif termination.startswith("equalize in"):
            self.result = EQUAL_IN
            self.moves = int(termination.split()[-1])
            self.cp = None
        elif "cp in" in termination:
            self.result = EVAL_IN
            parts = termination.split()
            self.moves = int(parts[-1])
            self.cp = int(parts[0][1:-2])
        elif "promotion with" in termination:
            self.result = PROMOTION
            self.moves = None
            parts = termination.split()
            self.cp = int(parts[-1][1:-2])
        else:
            self.result = MATE
            self.moves = None
            self.cp = None
            # print("No termination tag, expecting MATE", termination)


class LearnModel(GameModel):
    __gsignals__ = {
        # goal_checked is emitted after puzzle game goal check finished
        "goal_checked": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "learn_success": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def isChanged(self):
        """Never save learn games changes to .pgn"""
        return False

    def set_learn_data(
        self, learn_type, source, current_index=None, game_count=None, from_lesson=False
    ):
        self.learn_type = learn_type
        self.source = source
        self.current_index = current_index
        self.game_count = game_count
        self.from_lesson = from_lesson

        self.hints = {}
        self.goal = None
        self.failed_playing_best = False

        if learn_type == LECTURE:
            self.offline_lecture = True
            self.lecture_skip_event = asyncio.Event()  # set when 'Go on' button pressed
            self.lecture_pause_event = (
                asyncio.Event()
            )  # set when 'Pause' button pressed
            self.lecture_exit_event = asyncio.Event()  # set when 'Exit' button pressed

        elif learn_type == PUZZLE:
            self.puzzle_game = True
            self.goal = Goal(self.tags["Termination"])

        elif learn_type == LESSON:
            self.lesson_game = True

        elif learn_type == ENDGAME:
            self.end_game = True

    def add_hints(self, hint):
        ply, move = hint
        self.hints[ply] = [(move, 10000)]

    def check_failed_playing_best(self, status):
        if self.ply - 1 in self.hints:
            best_score = self.hints[self.ply - 1][0][1]
            best_moves = [
                hint[0]
                for hint in self.hints[self.ply - 1]
                if abs(hint[1] - best_score) <= 10
            ]
        else:
            best_moves = []

        if self.puzzle_game:
            # No need to check in best moves (and let add more time to analyzer) in trivial cases
            if self.goal.result in (MATE, MATE_IN) and (
                (status == BLACKWON and self.starting_color == BLACK)
                or (status == WHITEWON and self.starting_color == WHITE)
            ):
                return False
            elif self.goal.result == DRAW_IN and status == DRAW:
                return False
            elif self.goal.result in (MATE_IN, DRAW_IN, EQUAL_IN, EVAL_IN, PROMOTION):
                expect_best = True
            else:
                expect_best = False
        else:
            expect_best = False

        failed = expect_best and self.moves[-1].as_uci() not in best_moves
        return failed

    def check_goal(self, status, reason):
        if self.end_game:
            if status in UNDOABLE_STATES:
                self.end(status, reason)
            self.emit("goal_checked")
            return

        full_moves = (self.ply - self.lowply) // 2 + 1
        # print("Is Goal not reached?", self.goal.result, status, full_moves, self.goal.moves, self.failed_playing_best, self.tags, self.hints)

        if (
            (
                self.goal.result == DRAW_IN
                and status == DRAW
                and full_moves <= self.goal.moves
            )
            or (
                self.goal.result == MATE_IN
                and status == WHITEWON
                and full_moves <= self.goal.moves
                and self.starting_color == WHITE
            )
            or (
                self.goal.result == MATE_IN
                and status == BLACKWON
                and full_moves <= self.goal.moves
                and self.starting_color == BLACK
            )
            or (
                self.goal.result in (EVAL_IN, EQUAL_IN)
                and full_moves == self.goal.moves
                and not self.failed_playing_best
            )
            or (self.goal.result == MATE and status in (WHITEWON, BLACKWON))
            or (self.goal.result == PROMOTION and self.moves[-1].promotion)
        ):
            if status in UNDOABLE_STATES:
                self.end(status, PRACTICE_GOAL_REACHED)
            else:
                self.end(CANCELLED, PRACTICE_GOAL_REACHED)
        else:
            if status in UNDOABLE_STATES:
                # print("check_goal() status in UNDOABLE_STATES -> self.end(status, reason)")
                self.failed_playing_best = True
                self.end(status, reason)
            # print("Goal not reached yet.", self.goal.result, status, full_moves, self.goal.moves, self.failed_playing_best)

        self.emit("goal_checked")
        return



import os
import sys
import shutil

from gi.repository import Gtk, Gdk, GLib, GObject, Pango
from gi.repository.GdkPixbuf import Pixbuf

from pychess.System import uistuff
from pychess.System.prefix import getEngineDataPrefix, addDataPrefix
from pychess.Utils.IconLoader import get_pixbuf
from pychess.Players.engineNest import discoverer, is_uci, is_cecp, ENGINE_DEFAULT_LEVEL
from pychess.Players.engineList import VM_LIST, listEnginesFromPath
from pychess.Utils.isoCountries import ISO3166_LIST
from pychess.widgets import newGameDialog
from pychess.widgets import mainwindow

firstRun = True
engine_dialog = None


def run(widgets):
    global firstRun, engine_dialog
    if firstRun:
        # Display of the countries
        items = []
        for iso in ISO3166_LIST:
            path = addDataPrefix("flags/%s.png" % iso.iso2)
            if not (iso.iso2 and os.path.isfile(path)):
                path = addDataPrefix("flags/unknown.png")
            items.append((get_pixbuf(path), iso.country))
        uistuff.createCombo(
            widgets["engine_country_combo"],
            name="engine_country_combo",
            ellipsize_mode=Pango.EllipsizeMode.END,
        )
        data = [(item[0], item[1]) for item in items]
        uistuff.updateCombo(widgets["engine_country_combo"], data)

        engine_dialog = EnginesDialog(widgets)

        def cancel_event(widget, with_confirmation, *args):
            # Confirm if the changes need to be saved
            modified = discoverer.hasChanged()
            if modified and with_confirmation:
                dialog = Gtk.MessageDialog(
                    mainwindow(),
                    type=Gtk.MessageType.QUESTION,
                    buttons=Gtk.ButtonsType.YES_NO,
                )
                dialog.set_markup(
                    _("You have unsaved changes. Do you want to save before leaving?")
                )
                response = dialog.run()
                dialog.destroy()
                # if response == Gtk.ResponseType.CANCEL:
                #    return False
                if response == Gtk.ResponseType.NO:
                    discoverer.restore()
                if response == Gtk.ResponseType.YES:
                    discoverer.save()

            # Close the window
            widgets["manage_engines_dialog"].hide()
            return True

        def save_event(widget, *args):
            discoverer.save()
            widgets["manage_engines_dialog"].hide()
            return True

        widgets["manage_engines_dialog"].connect("delete-event", cancel_event, True)
        widgets["engine_cancel_button"].connect("clicked", cancel_event, False)
        widgets["engine_save_button"].connect("clicked", save_event)
        widgets["manage_engines_dialog"].connect(
            "key-press-event",
            lambda w, e: cancel_event(w, True) if e.keyval == Gdk.KEY_Escape else None,
        )

    discoverer.backup()
    engine_dialog.widgets["enginebook"].set_current_page(0)
    widgets["manage_engines_dialog"].show()
    if not firstRun:
        engine_dialog.update_store()
    firstRun = False


class EnginesDialog:
    def update_options(self, *args):
        if self.cur_engine is not None:
            # Initial reset
            self.options_store.clear()

            # Detection of the name of the engine to reload
            engines = discoverer.getEngines()
            names = [engine["name"] for engine in engines]
            if self.cur_engine not in names:
                self.cur_engine = engines[0]["name"]
            engine = discoverer.getEngineByName(self.cur_engine)
            if engine:
                options = engine.get("options")
                if options:
                    options.sort(
                        key=lambda obj: obj["name"].lower() if "name" in obj else ""
                    )
                    for option in options:
                        key = option["name"]
                        val = option
                        if option["type"] != "button":
                            val["default"] = option.get("default")
                            val["value"] = option.get("value", val["default"])
                            modified = val["value"] != val["default"]
                        else:
                            modified = False
                        self.options_store.append(["*" if modified else "", key, val])

    def update_store(self, *args):
        newGameDialog.createPlayerUIGlobals(discoverer)
        engine_names = [row[1] for row in self.allstore]
        new_items = []
        # don't add the very first (Human) player to engine store
        for item in newGameDialog.allEngineItems:
            if item[1] not in engine_names:
                new_items.append(item)
        ts_iter = None
        for item in new_items:
            ts_iter = self.allstore.append(item)
        if ts_iter is not None:
            text_select = self.tv.get_selection()
            text_select.select_iter(ts_iter)
        self.update_options()

    def __init__(self, widgets):
        self.widgets = widgets
        self.dialog = self.widgets["manage_engines_dialog"]
        self.cur_engine = None
        self.default_workdir = getEngineDataPrefix()

        uistuff.keepWindowSize("engineswindow", self.dialog)

        # Put engines into tree store
        self.allstore = Gtk.ListStore(Pixbuf, str)

        self.tv = self.widgets["engines_treeview"]
        self.tv.set_model(self.allstore)
        self.tv.append_column(
            Gtk.TreeViewColumn("Flag", Gtk.CellRendererPixbuf(), pixbuf=0)
        )
        name_renderer = Gtk.CellRendererText()
        name_renderer.set_property("editable", False)
        self.tv.append_column(Gtk.TreeViewColumn("Name", name_renderer, text=1))

        # Add cell renderer to protocol combo column
        protocol_combo = self.widgets["engine_protocol_combo"]
        protocol_combo.set_name("engine_protocol_combo")
        cell = Gtk.CellRendererText()
        protocol_combo.pack_start(cell, True)
        protocol_combo.add_attribute(cell, "text", 0)

        # Add columns and cell renderers to options treeview
        self.options_store = Gtk.ListStore(str, str, GObject.TYPE_PYOBJECT)
        optv = self.widgets["options_treeview"]
        optv.set_model(self.options_store)
        optv.append_column(Gtk.TreeViewColumn("  ", Gtk.CellRendererText(), text=0))
        optv.append_column(
            Gtk.TreeViewColumn(_("Option"), Gtk.CellRendererText(), text=1)
        )
        optv.append_column(
            Gtk.TreeViewColumn(
                _("Value"), KeyValueCellRenderer(self.options_store), data=2
            )
        )

        self.update_store()

        def do_update_store(self, *args):
            GLib.idle_add(engine_dialog.update_store)

        discoverer.connect_after("engine_discovered", do_update_store)

        ################################################################
        # remove button
        ################################################################
        def remove(button):
            if self.cur_engine is not None:
                self.widgets["remove_engine_button"].set_sensitive(False)
                discoverer.removeEngine(self.cur_engine)

                selection = self.tv.get_selection()
                result = selection.get_selected()
                if result is not None:
                    model, ts_iter = result
                    model.remove(ts_iter)
                if model.iter_n_children() == 0:
                    clearView()

                # Notify playerCombos in NewGameTasker
                discoverer.emit("all_engines_discovered")

        self.widgets["remove_engine_button"].connect("clicked", remove)

        ################################################################
        # add button
        ################################################################
        engine_chooser_dialog = Gtk.FileChooserDialog(
            _("Select engine"),
            mainwindow(),
            Gtk.FileChooserAction.OPEN,
            (
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
            ),
        )

        filter = Gtk.FileFilter()
        filter.set_name(_("Executable files"))
        filter.add_mime_type("application/x-executable")
        filter.add_mime_type("application/x-sharedlib")
        filter.add_mime_type("application/x-ms-dos-executable")
        filter.add_mime_type("application/x-msdownload")
        filter.add_pattern("*.exe")
        for vm in VM_LIST:
            filter.add_pattern("*%s" % vm.ext)

        engine_chooser_dialog.add_filter(filter)
        self.add = False

        def add(button):
            self.add = True
            response = engine_chooser_dialog.run()

            if response == Gtk.ResponseType.OK:
                new_engine = engine_chooser_dialog.get_filename()
                binname = os.path.split(new_engine)[1]
                ext = os.path.splitext(new_engine)[1]

                # Verify if the engine already exists under the same name
                if new_engine != "":
                    for eng in discoverer.getEngines():
                        if eng["command"] == new_engine:
                            msg_dia = Gtk.MessageDialog(
                                mainwindow(),
                                type=Gtk.MessageType.ERROR,
                                buttons=Gtk.ButtonsType.OK,
                            )
                            msg_dia.set_markup(
                                _("<big><b>Unable to add %s</b></big>" % new_engine)
                            )
                            msg_dia.format_secondary_text(
                                _("The engine is already installed under the same name")
                            )
                            msg_dia.run()
                            msg_dia.hide()
                            new_engine = ""
                            break

                # Detect the host application
                if new_engine != "":
                    vm_name = None
                    vm_args = None
                    vmpath = ""

                    # Scripting
                    for vm in VM_LIST:
                        if ext == vm.ext:
                            vm_name = vm.name
                            vm_args = vm.args
                            break

                    # Wine for Windows application under Linux
                    if (
                        vm_name is None
                        and new_engine.lower().endswith(".exe")
                        and sys.platform != "win32"
                    ):
                        vm_name = "wine"

                    # Check that the interpreter is available
                    if vm_name is not None:
                        vmpath = shutil.which(vm_name, mode=os.R_OK | os.X_OK)
                        if vmpath is None:
                            msg_dia = Gtk.MessageDialog(
                                mainwindow(),
                                type=Gtk.MessageType.ERROR,
                                buttons=Gtk.ButtonsType.OK,
                            )
                            msg_dia.set_markup(
                                _("<big><b>Unable to add %s</b></big>" % new_engine)
                            )
                            msg_dia.format_secondary_text(
                                vm_name + _(" is not installed")
                            )
                            msg_dia.run()
                            msg_dia.hide()
                            new_engine = ""

                # Next checks
                if new_engine:
                    vm_ext_list = [vm.ext for vm in VM_LIST]
                    if ext not in vm_ext_list and not os.access(new_engine, os.X_OK):
                        msg_dia = Gtk.MessageDialog(
                            mainwindow(),
                            type=Gtk.MessageType.ERROR,
                            buttons=Gtk.ButtonsType.OK,
                        )
                        msg_dia.set_markup(
                            _(
                                "<big><b>%s is not marked executable in the filesystem</b></big>"
                                % new_engine
                            )
                        )
                        msg_dia.format_secondary_text(
                            _("Try chmod a+x %s" % new_engine)
                        )
                        msg_dia.run()
                        msg_dia.hide()
                        self.add = False
                        engine_chooser_dialog.hide()
                        return

                    try:
                        engine_command = []
                        if vmpath:
                            engine_command.append(vmpath)
                        if vm_args is not None:
                            engine_command += vm_args
                        engine_command.append(new_engine)

                        # Search the engines based on the most expectable protocol
                        refeng = discoverer.getReferencedEngine(binname)
                        if refeng is not None and refeng["protocol"] == "xboard":
                            checkers = [is_cecp, is_uci]
                        else:
                            checkers = [is_uci, is_cecp]

                        uci = False
                        for checker in checkers:
                            check_ok = checker(engine_command)
                            if check_ok:
                                uci = checker is is_uci
                                break

                        if not check_ok:
                            # restore the original
                            engine = discoverer.getEngineByName(self.cur_engine)
                            engine_chooser_dialog.set_filename(engine["command"])
                            msg_dia = Gtk.MessageDialog(
                                mainwindow(),
                                type=Gtk.MessageType.ERROR,
                                buttons=Gtk.ButtonsType.OK,
                            )
                            msg_dia.set_markup(
                                _("<big><b>Unable to add %s</b></big>" % new_engine)
                            )
                            msg_dia.format_secondary_text(
                                _("There is something wrong with this executable")
                            )
                            msg_dia.run()
                            msg_dia.hide()
                            engine_chooser_dialog.hide()
                            self.add = False
                            engine_chooser_dialog.hide()
                            return

                        self.widgets["engine_command_entry"].set_text(new_engine)
                        self.widgets["engine_protocol_combo"].set_active(
                            0 if uci else 1
                        )
                        self.widgets["engine_args_entry"].set_text("")

                        # active = self.widgets["engine_protocol_combo"].get_active()
                        protocol = "uci" if uci else "xboard"

                        # print(binname, new_engine, protocol, vm_name, vm_args)
                        discoverer.addEngine(
                            binname, new_engine, protocol, vm_name, vm_args
                        )
                        self.cur_engine = binname
                        self.add = False
                        discoverer.discover()
                    except Exception:
                        msg_dia = Gtk.MessageDialog(
                            mainwindow(),
                            type=Gtk.MessageType.ERROR,
                            buttons=Gtk.ButtonsType.OK,
                        )
                        msg_dia.set_markup(
                            _("<big><b>Unable to add %s</b></big>" % new_engine)
                        )
                        msg_dia.format_secondary_text(
                            _("There is something wrong with this executable")
                        )
                        msg_dia.run()
                        msg_dia.hide()
                        self.add = False
                        engine_chooser_dialog.hide()
                        return
                else:
                    # restore the original
                    engine = discoverer.getEngineByName(self.cur_engine)
                    engine_chooser_dialog.set_filename(engine["command"])

            engine_chooser_dialog.hide()

        self.widgets["add_engine_button"].connect("clicked", add)

        ################################################################
        # add in mass button
        ################################################################

        def addInMass(button):
            # Ask the user to select a folder
            folder_dlg = Gtk.FileChooserDialog(
                _("Choose a folder"),
                None,
                Gtk.FileChooserAction.SELECT_FOLDER,
                (
                    Gtk.STOCK_CANCEL,
                    Gtk.ResponseType.CANCEL,
                    Gtk.STOCK_OPEN,
                    Gtk.ResponseType.OK,
                ),
            )
            answer = folder_dlg.run()
            path = folder_dlg.get_filename()
            folder_dlg.destroy()

            # Search for the engines
            if answer != Gtk.ResponseType.OK:
                return False
            possibleFiles = listEnginesFromPath(path)

            # Remove the existing engines from the list
            def isNewEngine(path):
                sfn = os.path.basename(path)
                for engine in discoverer.getEngines():
                    if sfn in engine.get("command"):  # The short name must be unique
                        return False
                return True

            possibleFiles = [fn for fn in possibleFiles if isNewEngine(fn)]
            if len(possibleFiles) == 0:
                return False

            # Prepare the result in a dialog
            mass_dialog = self.widgets["engine_list_dialog"]
            self.widgets["mass_path_label"].set_text(path)
            mass_list = self.widgets["mass_list_treeview"]
            if len(mass_list.get_columns()) == 0:  # Not initialized yet
                mass_store = Gtk.ListStore(bool, str)
                mass_list.set_model(mass_store)

                def checkbox_renderer_cb(cell, path, model):
                    model[path][0] = not model[path][0]
                    return

                checkbox_renderer = Gtk.CellRendererToggle()
                checkbox_renderer.set_property("activatable", True)
                checkbox_renderer.connect("toggled", checkbox_renderer_cb, mass_store)
                mass_list.append_column(
                    Gtk.TreeViewColumn(_("Import"), checkbox_renderer, active=0)
                )
                mass_list.append_column(
                    Gtk.TreeViewColumn(_("File name"), Gtk.CellRendererText(), text=1)
                )
            else:
                mass_store = mass_list.get_model()

            mass_store.clear()
            for fn in possibleFiles:
                mass_store.append([False, fn[len(path) :]])

            # Show the dialog
            answer = mass_dialog.run()
            mass_dialog.hide()
            if answer != Gtk.ResponseType.OK.real:
                return False

            # Add the new engines
            self.add = True
            found = False
            for entry in mass_store:
                if entry[0]:
                    newengine = discoverer.getReferencedEngine(path + entry[1])
                    if newengine is not None:
                        discoverer.addEngineFromReference(newengine)
                        found = True
            self.add = False
            if found:
                discoverer.discover()
            return True

        self.widgets["mass_engine_button"].connect("clicked", addInMass)

        ################################################################
        def clearView():
            self.selection = True
            self.cur_engine = None
            self.widgets["vm_command_entry"].set_text("")
            self.widgets["vm_args_entry"].set_text("")
            self.widgets["engine_command_entry"].set_text("")
            self.widgets["engine_args_entry"].set_text("")
            self.widgets["engine_protocol_combo"].set_active(0)
            self.widgets["engine_country_combo"].set_active(0)
            self.widgets["engine_comment_entry"].set_text("")
            self.widgets["engine_level_scale"].set_value(ENGINE_DEFAULT_LEVEL)
            self.options_store.clear()
            self.selection = False

        ################################################################
        # vm args
        ################################################################
        def vm_args_changed(widget):
            if self.cur_engine is not None and not self.selection:
                new_args = self.widgets["vm_args_entry"].get_text().strip()
                engine = discoverer.getEngineByName(self.cur_engine)
                old_args = engine.get("vm_args")
                if new_args != old_args:
                    engine["vm_args"] = new_args.split()

        self.widgets["vm_args_entry"].connect("changed", vm_args_changed)

        ################################################################
        # engine args
        ################################################################
        def args_changed(widget):
            if self.cur_engine is not None and not self.selection:
                new_args = self.widgets["engine_args_entry"].get_text().strip()
                engine = discoverer.getEngineByName(self.cur_engine)
                old_args = engine.get("args")
                if new_args != old_args:
                    engine["args"] = new_args.split()

        self.widgets["engine_args_entry"].connect("changed", args_changed)

        ################################################################
        # engine working directory
        ################################################################
        dir_chooser_dialog = Gtk.FileChooserDialog(
            _("Select working directory"),
            mainwindow(),
            Gtk.FileChooserAction.SELECT_FOLDER,
            (
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
            ),
        )
        dir_chooser_button = Gtk.FileChooserButton.new_with_dialog(dir_chooser_dialog)

        self.widgets["dirChooserDock"].add(dir_chooser_button)
        dir_chooser_button.show()

        def select_dir(button):
            new_directory = dir_chooser_dialog.get_filename()
            engine = discoverer.getEngineByName(self.cur_engine)
            old_directory = engine.get("workingDirectory")
            if new_directory != old_directory and new_directory != self.default_workdir:
                engine["workingDirectory"] = new_directory

        dir_chooser_button.connect("current-folder-changed", select_dir)

        ################################################################
        # engine protocol
        ################################################################
        def protocol_changed(widget):
            if self.cur_engine is not None and not self.add and not self.selection:
                active = self.widgets["engine_protocol_combo"].get_active()
                new_protocol = "uci" if active == 0 else "xboard"
                engine = discoverer.getEngineByName(self.cur_engine)
                old_protocol = engine["protocol"]
                if new_protocol != old_protocol:
                    command = engine.get("command")
                    engine_command = []
                    vm_command = engine.get("vm_command")
                    if vm_command is not None:
                        engine_command.append(vm_command)
                        vm_args = engine.get("vm_args")
                        if vm_args is not None:
                            engine_command.append(", ".join(vm_args))
                    engine_command.append(command)

                    # is the new protocol supported by the engine?
                    if new_protocol == "uci":
                        check_ok = is_uci(engine_command)
                    else:
                        check_ok = is_cecp(engine_command)

                    if check_ok:
                        # discover engine options for new protocol
                        engine["protocol"] = new_protocol
                        engine["recheck"] = True
                        discoverer.discover()
                    else:
                        # restore the original protocol
                        widgets["engine_protocol_combo"].set_active(
                            0 if old_protocol == "uci" else 1
                        )

        self.widgets["engine_protocol_combo"].connect("changed", protocol_changed)

        ################################################################
        # engine country
        ################################################################
        def country_changed(widget):
            if self.cur_engine is not None and not self.selection:
                engine = discoverer.getEngineByName(self.cur_engine)
                old_country = discoverer.getCountry(engine)
                new_country = ISO3166_LIST[widget.get_active()].iso2
                if old_country != new_country:
                    engine["country"] = new_country

                    # Refresh the flag in the tree view
                    path = addDataPrefix("flags/%s.png" % new_country)
                    if not os.path.isfile(path):
                        path = addDataPrefix("flags/unknown.png")
                    item = self.tv.get_selection().get_selected()
                    if item is not None:
                        model, ts_iter = item
                        model[ts_iter][0] = get_pixbuf(path)

                        # Notify playerCombos in NewGameTasker
                        discoverer.emit("all_engines_discovered")

        self.widgets["engine_country_combo"].connect("changed", country_changed)

        def country_keypressed(widget, event):
            idx = 0
            for iso in ISO3166_LIST:
                if (idx != 0) and (
                    (ord(iso.country[0].lower()) == event.keyval)
                    or (ord(iso.country[0].upper()) == event.keyval)
                ):
                    widget.set_active(idx)
                    break
                idx += 1

        self.widgets["engine_country_combo"].connect(
            "key-press-event", country_keypressed
        )

        ################################################################
        # comment changed
        ################################################################
        def comment_changed(widget):
            if self.cur_engine is not None and not self.selection:
                new_comment = self.widgets["engine_comment_entry"].get_text().strip()
                engine = discoverer.getEngineByName(self.cur_engine)
                old_comment = engine.get("comment")
                if new_comment != old_comment:
                    engine["comment"] = new_comment

        self.widgets["engine_comment_entry"].connect("changed", comment_changed)

        ################################################################
        # level changed
        ################################################################
        def level_changed(widget):
            if self.cur_engine is not None and not self.selection:
                new_level = widget.get_value()
                engine = discoverer.getEngineByName(self.cur_engine)
                old_level = engine.get("level")
                if new_level != old_level:
                    engine["level"] = int(new_level)

        self.widgets["engine_level_scale"].connect("value-changed", level_changed)

        ################################################################
        # engine tree
        ################################################################
        self.selection = False

        def selection_changed(treeselection):
            store, tv_iter = self.tv.get_selection().get_selected()
            if tv_iter:
                self.selection = True
                path = store.get_path(tv_iter)
                indices = path.get_indices()
                row = indices[0]
                name = store[row][1]
                self.cur_engine = name
                engine = discoverer.getEngineByName(name)
                if "PyChess.py" in engine["command"]:
                    self.widgets["remove_engine_button"].set_sensitive(False)
                else:
                    self.widgets["remove_engine_button"].set_sensitive(True)
                self.widgets["engine_command_entry"].set_text(engine["command"])
                engine_chooser_dialog.set_filename(engine["command"])
                args = [] if engine.get("args") is None else engine.get("args")
                self.widgets["engine_args_entry"].set_text(" ".join(args))

                vm = engine.get("vm_command")
                self.widgets["vm_command_entry"].set_text(vm if vm is not None else "")
                args = [] if engine.get("vm_args") is None else engine.get("vm_args")
                self.widgets["vm_args_entry"].set_text(" ".join(args))

                directory = engine.get("workingDirectory")
                dir_choice = (
                    directory if directory is not None else self.default_workdir
                )
                dir_chooser_dialog.set_current_folder(dir_choice)
                self.widgets["engine_protocol_combo"].set_active(
                    0 if engine["protocol"] == "uci" else 1
                )

                self.widgets["engine_country_combo"].set_active(0)
                country = discoverer.getCountry(engine)
                idx = 0
                for iso in ISO3166_LIST:
                    if iso.iso2 == country:
                        self.widgets["engine_country_combo"].set_active(idx)
                        break
                    idx += 1

                comment = engine.get("comment")
                self.widgets["engine_comment_entry"].set_text(
                    comment if comment is not None else ""
                )

                level = engine.get("level")
                try:
                    level = int(level)
                except Exception:
                    level = ENGINE_DEFAULT_LEVEL
                self.widgets["engine_level_scale"].set_value(level)

                self.update_options()
                self.selection = False

        tree_selection = self.tv.get_selection()
        tree_selection.connect("changed", selection_changed)
        tree_selection.select_path((0,))
        selection_changed(tree_selection)

        ################################################################
        # restore the default options of the engine
        ################################################################

        def engine_default_options(button):
            if self.cur_engine is not None and not self.selection:
                engine = discoverer.getEngineByName(self.cur_engine)
                options = engine.get("options")
                if options:
                    dialog = Gtk.MessageDialog(
                        mainwindow(),
                        type=Gtk.MessageType.QUESTION,
                        buttons=Gtk.ButtonsType.YES_NO,
                    )
                    dialog.set_markup(
                        _(
                            "Do you really want to restore the default options of the engine ?"
                        )
                    )
                    response = dialog.run()
                    dialog.destroy()
                    if response == Gtk.ResponseType.YES:
                        for option in options:
                            if "default" in option:
                                option["value"] = option["default"]
                        self.update_options()

        self.widgets["engine_default_options_button"].connect(
            "clicked", engine_default_options
        )


class KeyValueCellRenderer(Gtk.CellRenderer):
    """ Custom renderer providing different renderers in different rows.
        The model parameter is a Gtk.ListStore(str, GObject.TYPE_PYOBJECT)
        containing key data pairs. Each data is a dictionary with
        name, type, default, value, min, max (for spin options), choices (list of combo options)
        The 'type' can be 'check', 'spin', 'text', 'combo', 'button'.
        Examples:
            ('Nullmove', {'name': 'Nullmove', 'default': false, 'type': 'check', 'value': True})
            ('Selectivity', {'name': 'Selectivity', 'default': 1, 'type': 'spin', \
                'min': 0, 'max': 4, 'value': 2})
            ('Style', {'name': 'Style', 'default': 'Solid', 'type': 'combo', \
                'choices': ['Solid', 'Normal','Risky'], 'value': 'Normal'})
            ('NalimovPath', {'name': 'NalimovPath', 'default': '',  \
                'type': 'text', 'value': '/home/egtb'})
            ('Clear Hash', {'name': 'Clear Hash', 'type': 'button'})
    """

    __gproperties__ = {
        "data": (
            GObject.TYPE_PYOBJECT,
            "Data",
            "Data",
            GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
        )
    }

    def __init__(self, model):
        GObject.GObject.__init__(self)
        self.data = None

        self.text_renderer = Gtk.CellRendererText()
        self.text_renderer.set_property("editable", True)
        self.text_renderer.connect("edited", self.text_edited_cb, model)

        self.toggle_renderer = Gtk.CellRendererToggle()
        self.toggle_renderer.set_property("activatable", True)
        self.toggle_renderer.set_property("xalign", 0)
        self.toggle_renderer.connect("toggled", self.toggled_cb, model)

        self.ro_toggle_renderer = Gtk.CellRendererToggle()
        self.ro_toggle_renderer.set_property("activatable", False)
        self.ro_toggle_renderer.set_property("xalign", 0)

        self.spin_renderer = Gtk.CellRendererSpin()
        self.spin_renderer.set_property("editable", True)
        self.spin_renderer.connect("edited", self.spin_edited_cb, model)

        self.combo_renderer = Gtk.CellRendererCombo()
        self.combo_renderer.set_property("has_entry", False)
        self.combo_renderer.set_property("editable", True)
        self.combo_renderer.set_property("text_column", 0)
        self.combo_renderer.connect("edited", self.text_edited_cb, model)

        self.button_renderer = Gtk.CellRendererText()
        self.button_renderer.set_property("editable", False)

    def update_modified_mark(self, ref):
        ref[0] = "*" if ref[2].get("value") != ref[2].get("default") else ""

    def text_edited_cb(self, cell, path, new_text, model):
        model[path][2]["value"] = new_text
        self.update_modified_mark(model[path])
        return

    def toggled_cb(self, cell, path, model):
        model[path][2]["value"] = not model[path][2]["value"]
        self.update_modified_mark(model[path])
        return

    def spin_edited_cb(self, cell, path, new_text, model):
        try:
            model[path][2]["value"] = int(new_text)
            self.update_modified_mark(model[path])
        except Exception:
            pass
        return

    def _get_renderer(self):
        if self.data["type"] == "check":
            if self.data["name"] == "UCI_Chess960":
                return self.ro_toggle_renderer
            else:
                return self.toggle_renderer
        elif self.data["type"] == "spin":
            return self.spin_renderer
        elif self.data["type"] == "text":
            return self.text_renderer
        elif self.data["type"] == "combo":
            return self.combo_renderer
        elif self.data["type"] == "button":
            return self.button_renderer

    renderer = property(_get_renderer)

    def do_set_property(self, pspec, value):
        if value["type"] == "check":
            self.toggle_renderer.set_active(value["value"])
            self.set_property("mode", Gtk.CellRendererMode.ACTIVATABLE)
        elif value["type"] == "spin":
            adjustment = Gtk.Adjustment(
                value=int(value["value"]),
                lower=value["min"],
                upper=value["max"],
                step_increment=1,
            )
            self.spin_renderer.set_property("adjustment", adjustment)
            self.spin_renderer.set_property("text", str(value["value"]))
            self.set_property("mode", Gtk.CellRendererMode.EDITABLE)
        elif value["type"] == "text":
            self.text_renderer.set_property("text", value["value"])
            self.set_property("mode", Gtk.CellRendererMode.EDITABLE)
        elif value["type"] == "combo":
            liststore = Gtk.ListStore(str)
            for choice in value["choices"]:
                liststore.append([choice])
            self.combo_renderer.set_property("model", liststore)
            self.combo_renderer.set_property("text", value["value"])
            self.set_property("mode", Gtk.CellRendererMode.EDITABLE)
        elif value["type"] == "button":
            self.button_renderer.set_property("text", "")

        setattr(self, pspec.name, value)

    def do_get_property(self, pspec):
        return getattr(self, pspec.name)

    def do_get_size(self, widget, cell_area=None):
        return self.renderer.get_size(widget, cell_area=cell_area)

    def do_render(self, ctx, widget, background_area, cell_area, flags):
        self.renderer.render(ctx, widget, background_area, cell_area, flags)

    def do_activate(self, event, widget, path, background_area, cell_area, flags):
        return self.renderer.activate(
            event, widget, path, background_area, cell_area, flags
        )

    def do_start_editing(self, event, widget, path, background_area, cell_area, flags):
        return self.renderer.start_editing(
            event, widget, path, background_area, cell_area, flags
        )


GObject.type_register(KeyValueCellRenderer)



"""Knightodds variant"""

from pychess.Utils.const import KNIGHTODDSCHESS, VARIANTS_ODDS
from pychess.Utils.Board import Board

KNIGHTODDSSTART = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/R1BQKBNR w KQkq - 0 1"


class KnightOddsBoard(Board):
    """:Description: Knight Odds variant plays with the same rules as normal chess
    but one side start the game with a knight missing
    """

    variant = KNIGHTODDSCHESS
    __desc__ = _("One player starts with one less knight piece")
    name = _("Knight odds")
    cecp_name = "normal"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_ODDS

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=KNIGHTODDSSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



from gi.repository import Gtk, GObject

from pychess.Utils.lutils.LBoard import LBoard
from pychess.Utils.lutils.lmove import toSAN, parseAN
from pychess.Utils.const import FEN_START
from pychess.System.prefix import addDataPrefix

__title__ = _("Openings")

__icon__ = addDataPrefix("glade/panel_book.svg")

__desc__ = _("Openings panel can filter game list by opening moves")


class OpeningTreePanel(Gtk.TreeView):
    def __init__(self, persp):
        GObject.GObject.__init__(self)
        self.persp = persp
        self.filtered = False

        self.persp.connect("chessfile_imported", self.on_chessfile_imported)

        self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.liststore = Gtk.ListStore(int, str, int, int)
        self.modelsort = Gtk.TreeModelSort(self.liststore)

        self.modelsort.set_sort_column_id(2, Gtk.SortType.DESCENDING)
        self.set_model(self.modelsort)

        self.set_headers_visible(True)

        column = Gtk.TreeViewColumn(_("Move"), Gtk.CellRendererText(), text=1)
        column.set_sort_column_id(1)
        column.connect("clicked", self.column_clicked, 1)
        self.append_column(column)

        column = Gtk.TreeViewColumn(_("Games"), Gtk.CellRendererText(), text=2)
        column.set_sort_column_id(2)
        column.connect("clicked", self.column_clicked, 2)
        self.append_column(column)

        column = Gtk.TreeViewColumn(_("Winning %"), Gtk.CellRendererProgress(), value=3)
        column.set_min_width(80)
        column.set_sort_column_id(3)
        column.connect("clicked", self.column_clicked, 3)
        self.append_column(column)

        self.conid = self.connect_after("row-activated", self.row_activated)

        self.board = LBoard()
        self.board.applyFen(FEN_START)

        self.columns_autosize()

        sw = Gtk.ScrolledWindow()
        sw.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        sw.add(self)

        self.box.pack_start(sw, True, True, 0)

        #  buttons
        toolbar = Gtk.Toolbar()

        firstButton = Gtk.ToolButton(stock_id=Gtk.STOCK_MEDIA_PREVIOUS)
        toolbar.insert(firstButton, -1)

        prevButton = Gtk.ToolButton(stock_id=Gtk.STOCK_MEDIA_REWIND)
        toolbar.insert(prevButton, -1)

        self.filterButton = Gtk.ToggleToolButton(Gtk.STOCK_FIND)
        self.filterButton.set_tooltip_text(_("Filter game list by opening moves"))
        toolbar.insert(self.filterButton, -1)

        firstButton.connect("clicked", self.on_first_clicked)
        prevButton.connect("clicked", self.on_prev_clicked)
        self.filterButton.connect("clicked", self.on_filter_clicked)

        tool_box = Gtk.Box()
        tool_box.pack_start(toolbar, False, False, 0)
        self.box.pack_start(tool_box, False, False, 0)

        self.box.show_all()

    def on_chessfile_imported(self, persp, chessfile):
        self.update_tree()

    def on_first_clicked(self, widget):
        while self.board.hist_move:
            self.board.popMove()
        self.update_tree()

    def on_prev_clicked(self, widget):
        if self.board.hist_move:
            self.board.popMove()
        self.update_tree()

    def on_filter_clicked(self, button):
        self.filtered = button.get_active()
        if not self.filtered:
            self.persp.filter_panel.filterButton.set_sensitive(True)
            self.filtered = True
            while self.board.hist_move:
                self.board.popMove()
            self.update_tree()
            self.filtered = False
        else:
            self.persp.filter_panel.filterButton.set_sensitive(False)
            self.update_tree()

    def column_clicked(self, col, data):
        self.set_search_column(data)

    def row_activated(self, widget, path, col):
        lmove = self.liststore[self.modelsort.convert_path_to_child_path(path)[0]][0]
        self.board.applyMove(lmove)
        self.update_tree()

    def update_tree(self, load_games=True):
        self.persp.gamelist.ply = self.board.plyCount
        if load_games and self.filtered:
            self.persp.chessfile.set_fen_filter(self.board.asFen())
            self.persp.gamelist.load_games()

        result = self.persp.chessfile.get_book_moves(self.board.asFen())
        self.clear_tree()
        for move, count, white_won, blackwon, draw in result:
            lmove = parseAN(self.board, move)
            perf = 0 if not count else round((white_won * 100.0 + draw * 50.0) / count)
            self.liststore.append([lmove, toSAN(self.board, lmove), count, perf])

    def clear_tree(self):
        selection = self.get_selection()
        if self.conid is not None and selection.handler_is_connected(self.conid):
            with GObject.signal_handler_block(selection, self.conid):
                self.liststore.clear()
        else:
            self.liststore.clear()



from gi.repository import Gtk, GObject

from .__init__ import NORTH, EAST, SOUTH, WEST, CENTER, reprPos


class PyDockComposite(Gtk.Alignment):
    def __init__(self, position, perspective):
        GObject.GObject.__init__(self, xscale=1, yscale=1)

        if position == NORTH or position == SOUTH:
            paned = Gtk.VPaned()
        elif position == EAST or position == WEST:
            paned = Gtk.HPaned()
        self.position = position
        self.perspective = perspective
        self.paned = paned
        self.add(self.paned)
        self.paned.show()

    def _del(self):
        for component in self.getComponents():
            component._del()

    def __repr__(self):
        return "composite {} ({}, {})".format(
            reprPos[self.position],
            repr(self.paned.get_child1()),
            repr(self.paned.get_child2()),
        )

    def dock(self, widget, position, title, id):
        assert position != CENTER, "POSITION_CENTER only makes sense for leaves"
        parent = self.get_parent()
        while not isinstance(parent, PyDockComposite):
            parent = parent.get_parent()
        from .PyDockLeaf import PyDockLeaf

        leaf = PyDockLeaf(widget, title, id, self.perspective)
        new = PyDockComposite(position, self.perspective)
        parent.changeComponent(self, new)
        new.initChildren(self, leaf)
        return leaf

    def changeComponent(self, old, new):
        if old == self.paned.get_child1():
            self.paned.remove(old)
            self.paned.pack1(new, resize=True, shrink=True)
        else:
            self.paned.remove(old)
            self.paned.pack2(new, resize=True, shrink=True)
        new.show()

    def removeComponent(self, component):
        if component == self.paned.get_child1():
            new = self.paned.get_child2()
        else:
            new = self.paned.get_child1()
        self.paned.remove(new)
        parent = self.get_parent()
        while not isinstance(parent, PyDockComposite):
            parent = parent.get_parent()
        parent.changeComponent(self, new)
        component._del()  # TODO: is this necessary?
        new.show()

    def getComponents(self):
        return self.paned.get_children()

    def initChildren(self, old, new, preserve_dimensions=False):
        if self.position == NORTH or self.position == WEST:
            self.paned.pack1(new, resize=True, shrink=True)
            self.paned.pack2(old, resize=True, shrink=True)
        elif self.position == SOUTH or self.position == EAST:
            self.paned.pack1(old, resize=True, shrink=True)
            self.paned.pack2(new, resize=True, shrink=True)
        old.show()
        new.show()

        def cb(widget, allocation):
            # Set initial position of the divider between the two panes of Gtk.Paned
            if allocation.height != 1:
                if self.position == NORTH:
                    pos = 0.381966011 * allocation.height
                elif self.position == SOUTH:
                    pos = 0.618033989 * allocation.height
                elif self.position == WEST:
                    pos = 0.381966011 * allocation.width
                elif self.position == EAST:
                    pos = 0.618033989 * allocation.width

                widget.set_position(int(pos + 0.5))
                widget.disconnect(conid)

        if not preserve_dimensions:
            conid = self.paned.connect("size-allocate", cb)

    def getPosition(self):
        """Returns NORTH or SOUTH if the children are packed vertically.
        Returns WEST or EAST if the children are packed horizontally.
        Returns CENTER if there is only one child"""
        return self.position



import asyncio

from gi.repository import Gtk

from pychess.Utils.const import HINT, SPY, BLACK, WHITE
from pychess.System import conf
from pychess.System import uistuff
from pychess.System.Log import log
from pychess.Utils import prettyPrintScore
from pychess.Utils.Move import listToMoves, parseAny
from pychess.Utils.lutils.lmove import ParsingError
from pychess.Players.engineNest import discoverer
from pychess.widgets.preferencesDialog import anal_combo_get_value, anal_combo_set_value
from pychess.widgets.InfoBar import InfoBarMessage, InfoBarMessageButton
from pychess.widgets import mainwindow
from pychess.widgets import InfoBar
from pychess.perspectives import perspective_manager


class AnalyzeGameDialog:
    def __init__(self):
        self.widgets = uistuff.GladeWidgets("analyze_game.glade")
        self.widgets["analyze_game"].set_transient_for(mainwindow())
        self.stop_event = asyncio.Event()

        uistuff.keep(self.widgets["fromCurrent"], "fromCurrent")
        uistuff.keep(self.widgets["shouldBlack"], "shouldBlack")
        uistuff.keep(self.widgets["shouldWhite"], "shouldWhite")
        uistuff.keep(self.widgets["threatPV"], "threatPV")
        uistuff.keep(self.widgets["showEval"], "showEval")
        uistuff.keep(self.widgets["showBlunder"], "showBlunder")
        uistuff.keep(self.widgets["max_analysis_spin"], "max_analysis_spin")
        uistuff.keep(
            self.widgets["variation_threshold_spin"], "variation_threshold_spin"
        )

        # Analyzing engines
        uistuff.createCombo(self.widgets["ana_combobox"], name="ana_combobox")

        from pychess.widgets import newGameDialog

        def update_analyzers_store(discoverer):
            data = [(item[0], item[1]) for item in newGameDialog.analyzerItems]
            uistuff.updateCombo(self.widgets["ana_combobox"], data)

        discoverer.connect_after("all_engines_discovered", update_analyzers_store)
        update_analyzers_store(discoverer)

        uistuff.keep(
            self.widgets["ana_combobox"],
            "ana_combobox",
            anal_combo_get_value,
            lambda combobox, value: anal_combo_set_value(
                combobox, value, "hint_mode", HINT
            ),
        )

        def hide_window(button, *args):
            self.widgets["analyze_game"].destroy()

        def abort():
            self.analyzer.pause()
            if self.threat_PV:
                self.inv_analyzer.pause()
            self.stop_event.set()
            self.widgets["analyze_game"].destroy()

        def run_analyze(button, *args):
            async def coro():
                persp = perspective_manager.get_perspective("games")
                gmwidg = persp.cur_gmwidg()
                gamemodel = gmwidg.gamemodel

                old_check_value = conf.get("analyzer_check")
                conf.set("analyzer_check", True)
                if HINT not in gamemodel.spectators:
                    try:
                        await asyncio.wait_for(gamemodel.start_analyzer(HINT), 5.0)
                    except asyncio.TimeoutError:
                        log.error("Got timeout error while starting hint analyzer")
                        return
                    except Exception:
                        log.error("Unknown error while starting hint analyzer")
                        return
                self.analyzer = gamemodel.spectators[HINT]
                gmwidg.menuitems["hint_mode"].active = True
                self.threat_PV = conf.get("ThreatPV")
                if self.threat_PV:
                    old_inv_check_value = conf.get("inv_analyzer_check")
                    conf.set("inv_analyzer_check", True)
                    if SPY not in gamemodel.spectators:
                        try:
                            await asyncio.wait_for(gamemodel.start_analyzer(SPY), 5.0)
                        except asyncio.TimeoutError:
                            log.error("Got timeout error while starting spy analyzer")
                            return
                        except Exception:
                            log.error("Unknown error while starting spy analyzer")
                            return
                    inv_analyzer = gamemodel.spectators[SPY]
                    gmwidg.menuitems["spy_mode"].active = True

                title = _("Game analyzing in progress...")
                text = _("Do you want to abort it?")
                content = InfoBar.get_message_content(
                    title, text, Gtk.STOCK_DIALOG_QUESTION
                )

                def response_cb(infobar, response, message):
                    conf.set("analyzer_check", old_check_value)
                    if self.threat_PV:
                        conf.set("inv_analyzer_check", old_inv_check_value)
                    message.dismiss()
                    abort()

                message = InfoBarMessage(Gtk.MessageType.QUESTION, content, response_cb)
                message.add_button(
                    InfoBarMessageButton(_("Abort"), Gtk.ResponseType.CANCEL)
                )
                gmwidg.replaceMessages(message)

                async def analyse_moves():
                    should_black = conf.get("shouldBlack")
                    should_white = conf.get("shouldWhite")
                    from_current = conf.get("fromCurrent")
                    start_ply = gmwidg.board.view.shown if from_current else 0
                    move_time = int(conf.get("max_analysis_spin"))
                    threshold = int(conf.get("variation_threshold_spin"))
                    for board in gamemodel.boards[start_ply:]:
                        if self.stop_event.is_set():
                            break

                        gmwidg.board.view.setShownBoard(board)
                        self.analyzer.setBoard(board)
                        if self.threat_PV:
                            inv_analyzer.setBoard(board)
                        await asyncio.sleep(move_time + 0.1)

                        ply = board.ply - gamemodel.lowply
                        color = (ply - 1) % 2
                        if (
                            ply - 1 in gamemodel.scores
                            and ply in gamemodel.scores
                            and (
                                (color == BLACK and should_black)
                                or (color == WHITE and should_white)
                            )
                        ):
                            oldmoves, oldscore, olddepth = gamemodel.scores[ply - 1]
                            oldscore = oldscore * -1 if color == BLACK else oldscore
                            score_str = prettyPrintScore(oldscore, olddepth)
                            moves, score, depth = gamemodel.scores[ply]
                            score = score * -1 if color == WHITE else score
                            diff = score - oldscore
                            if (
                                (diff > threshold and color == BLACK)
                                or (diff < -1 * threshold and color == WHITE)
                            ) and (
                                gamemodel.moves[ply - 1]
                                != parseAny(gamemodel.boards[ply - 1], oldmoves[0])
                            ):
                                if self.threat_PV:
                                    try:
                                        if ply - 1 in gamemodel.spy_scores:
                                            (
                                                oldmoves0,
                                                oldscore0,
                                                olddepth0,
                                            ) = gamemodel.spy_scores[ply - 1]
                                            score_str0 = prettyPrintScore(
                                                oldscore0, olddepth0
                                            )
                                            pv0 = listToMoves(
                                                gamemodel.boards[ply - 1],
                                                ["--"] + oldmoves0,
                                                validate=True,
                                            )
                                            if len(pv0) > 2:
                                                gamemodel.add_variation(
                                                    gamemodel.boards[ply - 1],
                                                    pv0,
                                                    comment="Threatening",
                                                    score=score_str0,
                                                    emit=False,
                                                )
                                    except ParsingError as e:
                                        # ParsingErrors may happen when parsing "old" lines from
                                        # analyzing engines, which haven't yet noticed their new tasks
                                        log.debug(
                                            "__parseLine: Ignored (%s) from analyzer: ParsingError%s"
                                            % (" ".join(oldmoves), e)
                                        )
                                try:
                                    pv = listToMoves(
                                        gamemodel.boards[ply - 1],
                                        oldmoves,
                                        validate=True,
                                    )
                                    gamemodel.add_variation(
                                        gamemodel.boards[ply - 1],
                                        pv,
                                        comment="Better is",
                                        score=score_str,
                                        emit=False,
                                    )
                                except ParsingError as e:
                                    # ParsingErrors may happen when parsing "old" lines from
                                    # analyzing engines, which haven't yet noticed their new tasks
                                    log.debug(
                                        "__parseLine: Ignored (%s) from analyzer: ParsingError%s"
                                        % (" ".join(oldmoves), e)
                                    )

                    self.widgets["analyze_game"].hide()
                    self.widgets["analyze_ok_button"].set_sensitive(True)
                    conf.set("analyzer_check", old_check_value)
                    if self.threat_PV:
                        conf.set("inv_analyzer_check", old_inv_check_value)
                    message.dismiss()

                    gamemodel.emit("analysis_finished")

                asyncio.create_task(analyse_moves())
                hide_window(None)

                return True

            asyncio.create_task(coro())

        self.widgets["analyze_game"].connect("delete-event", hide_window)
        self.widgets["analyze_cancel_button"].connect("clicked", hide_window)
        self.widgets["analyze_ok_button"].connect("clicked", run_analyze)

    def run(self):
        self.stop_event.clear()
        self.widgets["analyze_game"].show()
        self.widgets["analyze_game"].present()



from .ldata import BLACK_SQUARES
from pychess.Utils.const import (
    ASEAN_VARIANTS,
    WHITE,
    BLACK,
    KNIGHT,
    BISHOP,
    ROOK,
    QUEEN,
    PAWN,
)


def testFifty(board):
    if board.fifty >= 100:
        return True
    return False


drawSet = {
    (0, 0, 0, 0, 0, 0, 0, 0),  # KK
    (0, 1, 0, 0, 0, 0, 0, 0),  # KBK
    (1, 0, 0, 0, 0, 0, 0, 0),  # KNK
    (0, 0, 0, 0, 0, 1, 0, 0),  # KKB
    (0, 0, 0, 0, 1, 0, 0, 0),  # KNK
    (1, 0, 0, 0, 0, 1, 0, 0),  # KNKB
    (0, 1, 0, 0, 1, 0, 0, 0),  # KBKN
}

# Contains not 100% sure ones
drawSet2 = {
    (2, 0, 0, 0, 0, 0, 0, 0),  # KNNK
    (0, 0, 0, 0, 2, 0, 0, 0),  # KKNN
    (2, 0, 0, 0, 1, 0, 0, 0),  # KNNKN
    (1, 0, 0, 0, 2, 0, 0, 0),  # KNKNN
    (2, 0, 0, 0, 0, 1, 0, 0),  # KNNKB
    (0, 1, 0, 0, 2, 0, 0, 0),  # KBKNN
    (2, 0, 0, 0, 0, 0, 1, 0),  # KNNKR
    (0, 0, 1, 0, 2, 0, 0, 0),  # KRKNN
}


def testMaterial(board):
    """Tests if no players are able to win the game from the current
    position"""

    whitePieceCount = board.pieceCount[WHITE]
    blackPieceCount = board.pieceCount[BLACK]

    if whitePieceCount[PAWN] or blackPieceCount[PAWN]:
        return False

    if whitePieceCount[QUEEN] or blackPieceCount[QUEEN]:
        return False

    wn = whitePieceCount[KNIGHT]
    wb = whitePieceCount[BISHOP]
    wr = whitePieceCount[ROOK]
    bn = blackPieceCount[KNIGHT]
    bb = blackPieceCount[BISHOP]
    br = blackPieceCount[ROOK]

    if board.variant in ASEAN_VARIANTS and (wb or bb):
        return False

    if (wn, wb, wr, 0, bn, bb, br, 0) in drawSet:
        return True

    # Tests KBKB. Draw if bishops are of same color
    if not wn + wr + bn + br and wb == 1 and bb == 1:
        if (
            board.boards[WHITE][BISHOP] & BLACK_SQUARES
            != board.boards[BLACK][BISHOP] & BLACK_SQUARES
        ):
            return True


def testPlayerMatingMaterial(board, color):
    """Tests if given color has enough material to mate on board"""

    pieceCount = board.pieceCount[color]

    if (
        pieceCount[PAWN]
        or pieceCount[QUEEN]
        or pieceCount[ROOK]
        or (pieceCount[KNIGHT] + pieceCount[BISHOP] > 1)
    ):
        return True
    return False


# This could be expanded by the fruit kpk draw function, which can test if a
# certain king verus king and pawn posistion is winable.


def test(board):
    """Test if the position is drawn. Two-fold repetitions are counted."""
    return (
        board.repetitionCount(draw_threshold=2) > 1
        or testFifty(board)
        or testMaterial(board)
    )



"""Giveaway Variant"""

from pychess.Utils.const import GIVEAWAYCHESS
from .suicide import SuicideBoard


class GiveawayBoard(SuicideBoard):
    """:Description: This is the international version of Losing chess used on ICC as Giveaway and on Lichess as Antichess
    You must capture if you can, and the object is to lose all your pieces or to have no moves left.
    But in Giveaway, the king is just like any other piece.
    It can move into check and be captured, and you can even promote pawns to kings.
    """

    variant = GIVEAWAYCHESS
    __desc__ = _("ICC giveaway: https://www.chessclub.com/user/help/Giveaway")
    name = _("Giveaway")
    cecp_name = "giveaway"
